import json
from collections import UserDict
from typing import Callable, Self, Any, Iterable, Mapping

from pystrict import strict

from oldaplib.src.helpers.serializer import serializer


#@strict
@serializer
class ObservableDict(UserDict):
    __on_change: Callable[[Self], None]

    def __init__(self,
                 obj: Iterable | Mapping | None = None, *,
                 on_change: Callable[[Self], None] | None = None,
                 **kwargs):
        self.__on_change = on_change
        if obj:
            super().__init__(obj, **kwargs)
        else:
            super().__init__(**kwargs)

    def __setitem__(self, key, value):
        if self.__on_change:
            self.__on_change(self.copy())
        super().__setitem__(key, value)

    def __delitem__(self, key):
        if self.__on_change:
            self.__on_change(self.copy())
        super().__delitem__(key)


    def copy(self) -> Self:
        return ObservableDict(self.data.copy())

    def set_on_change(self, on_change: Callable[[Self], None]) -> None:
        self.__on_change = on_change

    def _as_dict(self):
        return self.data

    def changeset_clear(self):
        pass


if __name__ == "__main__":
    def cb(old: ObservableDict):
        print("---->CALLBACK", old)

    gugus = ObservableDict()
    if gugus:
        print("gugus TRUE")
    else:
        print("gugus FALSE")

    gaga = ObservableDict({'a': 'AA', 'b': 'BB'}, on_change=cb)
    print(gaga)
    jsonstr = json.dumps(gaga, default=serializer.encoder_default)
    print(jsonstr)
    gaga2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
    print(type(gaga2), gaga2)
    gaga2.set_on_change(cb)
    gaga2['xx'] = 'XXXXXX'
    print(gaga2)
    #gugusta: ObservableDict[str] = ObservableDict(cb)
    # gaga.update({'c': 'CC'})
    # print(gaga)
    # print(gaga.get('c'), gaga.get('x'))
    #
    # gugus = {'a': 'AAAA', 'b': 'BBBB'}
    # g = ObservableDict(cb, gugus)
    # print(g)
    # for k, v in gaga.items():
    #     print(k, '->', v)


"""
# Observalbe Set

This module implements a set which is observable. That means, that a callback function/method may
be given that is called whenever the set is changed (adding/removing of items)
"""
import json
from typing import List, Callable, Any, Iterable, Set, Self

from pystrict import strict

from oldaplib.src.enums.action import Action
from oldaplib.src.helpers.oldaperror import OldapErrorKey
from oldaplib.src.helpers.serializer import serializer


#@strict
@serializer
class ObservableSet(Set):
    """
    The ObservableSet class is a subclass of `Set` which allows the notification if the set is changed, that
    is items are added or removed. For this purpose, a callback function can be added to the set which is
    called whenever the set changes.
    """
    __on_change: Callable[[Self, Any], None]
    __on_change_data: Any

    def __init__(self,
                 setitems: Iterable | None = None,
                 on_change: Callable[[Self, Any], None] = None,
                 on_change_data: Any = None) -> None:
        """
        Constructor of the ObservableSet class

        :param setitems: The items the ObservableSet will be initialized with
        :param on_change: Callback function to be called when an item is added/removed
        :param on_change_data: data supplied to the callback function
        """
        self.__on_change = on_change
        self.__on_change_data = on_change_data
        if setitems is None:
            super().__init__(set())
        else:
            super().__init__(setitems)

    def __repr__(self) -> str:
        l = [repr(x) for x in self]
        return ", ".join(l)

    def __str__(self) -> str:
        return super().__str__()

    def __or__(self, other: Self) -> Self:
        if isinstance(other, ObservableSet):
            return ObservableSet(super().__or__(other))
        return NotImplemented

    def __ror__(self, other: Self) -> Self:
        pass

    def __ior__(self, other: Self) -> Self:
        tmp_copy = self.copy()
        super().__ior__(other)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)
        return self

    def __and__(self, other: Self) -> Self:
        return ObservableSet(super().__and__(other))

    def __iand__(self, other: Self) -> Self:
        tmp_copy = self.copy()
        super().__iand__(other)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)
        return self

    def __rsub__(self, other: Self) -> Self:
        pass

    def __sub__(self, other: Self) -> Self:
        return ObservableSet(super().__sub__(other))

    def __isub__(self, other: Self) -> Self:
        tmp_copy = self.copy()
        super().__isub__(other)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)
        return self


    def __eq__(self, other: Self) -> Self:
        return set(self) == set(other)

    def update(self, items: Iterable):
        tmp_copy = self.copy()
        super().update(items)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)


    def intersection_update(self, items: Iterable):
        tmp_copy = self.copy()
        super().intersection_update(items)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def difference_update(self, items: Iterable):
        tmp_copy = self.copy()
        super().difference_update(items)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def symmetric_difference_update(self, items: Iterable):
        tmp_copy = self.copy()
        super().symmetric_difference_update(items)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def add(self, item: Any) -> None:
        tmp_copy = self.copy()
        super().add(item)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def remove(self, item: Any) -> None:
        tmp_copy = self.copy()
        super().remove(item)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def discard(self, item: Any):
        tmp_copy = self.copy()
        super().discard(item)
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)
        return self

    def pop(self):
        tmp_copy = self.copy()
        len1 = len(self)
        super().pop()
        len2 = len(self)
        if len1 == len2:
            return
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def clear(self) -> None:
        tmp_copy = self.copy()
        super().clear()
        if self.__on_change is not None:
            self.__on_change(tmp_copy, self.__on_change_data)

    def copy(self) -> Self:
        return ObservableSet(super().copy(), on_change=self.__on_change)

    @property
    def toRdf(self) -> str:
        l = [x.toRdf if getattr(x, "toRdf", None) else x for x in self]
        return ", ".join(l)


    def _as_dict(self):
        return {'setitems': list(self)}

    def asSet(self):
        return super().copy()

    def on_change(self, func: Callable[[Self, Any], None], data: Any) -> None:
        self.__on_change = func
        self.__on_change_data = data

    def changeset_clear(self):
        pass




from enum import unique, Enum

from oldaplib.src.connection import Connection
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.project import Project
from oldaplib.src.resourceclass import ResourceClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_string import Xsd_string


@unique
class OldapListNodeAttr(Enum):
    """
    This enum class represents the fields used in the project model
    """
    OLDAPLISTNODE_IRI = 'oldap:oldapListNodeIri'  # virtual property, repents the RDF subject
    IN_SCHEME = 'skos:inScheme'
    BROADER_TRANSITIVE = 'skos:broaderTransitive'
    NEXT_NODE = 'oldap:nextNode'
    LEFT_INDEX = 'oldap:leftIndex'
    RIGHT_INDEX = 'oldap:rightIndex'
    NOTATION = 'skos:notation'
    NOTE = 'skos:note'
    PREF_LABEL = 'skos:prefLabel'
    ALT_LABEL = 'skos:altLabel'
    DEFINITION = 'skos:definition'
    CHANGE_NOTE = 'skos:changeNote'

class OldapListNode(Model):
    __datatypes = {
        OldapListNodeAttr.OLDAPLISTNODE_IRI: Iri,
        OldapListNodeAttr.IN_SCHEME: Iri,
        OldapListNodeAttr.BROADER_TRANSITIVE: Iri,
        OldapListNodeAttr.NEXT_NODE: Iri,
        OldapListNodeAttr.LEFT_INDEX: int,
        OldapListNodeAttr.RIGHT_INDEX: int,
        OldapListNodeAttr.NOTATION: Xsd_string,
        OldapListNodeAttr.NOTE: LangString,
        OldapListNodeAttr.PREF_LABEL: LangString,
        OldapListNodeAttr.ALT_LABEL: LangString,
        OldapListNodeAttr.DEFINITION: LangString,
        OldapListNodeAttr.CHANGE_NOTE: Xsd_string,
    }

    __system = None
    __listnodeclass = None

    def __init__(self, *,
                 con: IConnection,
                 **kwargs):
        super().__init__(connection=con)
        if self.__system is None:
            self.__system = Project.read(con, 'oldap')
        if self.__listnodeclass is None:
            self.__listnodeclass = ResourceClass.read(con, self.__system, Iri('oldap:OldapListNode'))
        print(self.__system)
        print(self.__listnodeclass)

if __name__ == '__main__':
    con = Connection(server='http://localhost:7200',
                     repo="oldap",
                     userId="rosenth",
                     credentials="RioGrande",
                     context_name="DEFAULT")
    oln = OldapListNode(con=con)



import re
from typing import Self, Type

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_base64Binary(Xsd):
    """
    Class that encodes and decodes binary data using the XML Scheme [xsd:base64Binary](https://www.w3.org/TR/xmlschema11-2/#base64Binary) datatype
    """

    __value: bytes

    def __init__(self, value: Self | bytes, validate: bool = True):
        """
        Constructor that encodes and decodes binary data using the XML Scheme xsd:base64Binary datatype
        :param value: Either another instance of Xsd_base64Binary or a bytes object
        :type value: Xsd_base64Binary | bytes
        :raises OldapErrorValue: If the value is not an instance of Xsd_base64Binary or a valid bytes object
        """
        if isinstance(value, Xsd_base64Binary):
            self.__value = value.__value
        elif isinstance(value, bytes):
            self.__value = value
        else:
            OldapErrorValue("Xsd_base64Binary requires bytes parameter")
        if validate:
            if len(value) % 4 != 0:
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:base64Binary.')
            if not bool(re.match(r'^[A-Za-z0-9+/]+={0,2}$', value.decode('utf-8'))):
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:base64Binary.')
            if not XsdValidator.validate(XsdDatatypes.base64Binary, value.decode('utf-8')):
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:base64Binary.')

    def __str__(self):
        """
        String representation based on utf-8 encoding
        :return: string
        """
        return self.__value.decode('utf-8')

    def __repr__(self):
        """
        String representation based on utf-8 encoding as used for constructor
        :return: string
        """
        return f'Xsd_base64Binary(b"{self.__value.decode('utf-8')}")'

    def __eq__(self, other: Self | None) -> bool:
        """
        Compare for equality
        :param other: Another Xsd_base64Binary instance
        :type other: Xsd_base64Binary | None
        :return: True or False
        """
        if other is None:
            return False
        return self.__value == other.__value

    def __hash__(self) -> int:
        """
        Return the hash value for the object
        :return: Hash value
        """
        return hash(self.__value)

    def _as_dict(self) -> dict[str, bytes]:
        """
        Used of JSON serialization with @serialisation decorator
        :return:
        """
        return {'value': self.__value}

    @classmethod
    def fromRdf(cls: Type['XsdBase64Binary'], value: str) -> Type['XsdBase64Binary']:
        """
        Converts an Xsd_base64Binary RDF string into an Xsd_base64Binary object
        :param value: RDF string
        :return: Instance of Xsd_base64Binary
        """
        return cls(value.encode('utf-8'), validate=False)

    @property
    def toRdf(self) -> str:
        """
        Converts an Xsd_base64Binary object to an RDF string
        :return: RDF string
        """
        return f'"{self.__value.decode('utf-8')}"^^xsd:base64Binary'

    @property
    def value(self) -> bytes:
        """
        Converts an Xsd_base64Binary object to a bytes object
        :return: bytes
        """
        return self.__value




from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_integer import Xsd_integer


#@strict
@serializer
class Xsd_nonNegativeInteger(Xsd_integer):
    """
    Implements the XML Schema [xsd:nonNegativeInteger](https://www.w3.org/TR/xmlschema11-2/#nonNegativeInteger) datatype. Inherits from Xsd_integer.
    """
    __value: int

    def __init__(self, value: Xsd_integer | int | str, validate: bool = True):
        """
        Constructor of the Xsd_nonNegativeInteger class.
        :param value: Another valid Xsd instance, a non-negative integer or a valid string representation.
        :type value: Xsd_integer | int | str
        :raises OldapErrorValue: If the value does not represent a valid non-negative integer.
        """
        super().__init__(value)
        if self._value < 0:
            raise OldapErrorValue('Value must be "0" or positive.')



import json
from typing import Iterator, Self, Iterable

from pystrict import strict

from oldaplib.src.dtypes.rdfset import RdfSet
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorType, OldapErrorKey
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd_string import Xsd_string


@serializer
class LanguageIn(RdfSet[Language], Notify):
    """
    This class implements the SHACL sh:languageIn datatype. It completely validates the input.
    If the validations failes, an OldapErrorValue is raised.
    """

    def __init__(self,
                 *args: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str,
                 value: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str | None = None):
        nargs = tuple()
        nvalue = None
        if len(args) == 0:
            if value is None:
                pass
            else:
                if isinstance(value, LanguageIn):
                    nvalue = value
                elif isinstance(value, (set, list, tuple)):
                    for v in value:
                        if not isinstance(v, (Language, str)):
                            raise OldapErrorType(f'Iterable contains element that is not an instance of "Language", but "{type(v).__name__}".')
                    try:
                        nvalue = [x if isinstance(x, Language) else Language[x.upper()] for x in value]
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
                else:
                    if not isinstance(value, (Language, str)):
                        raise OldapErrorType(f'Value is not an instance of "Language", but "{type(value).__name__}".')
                    try:
                        nvalue = value if isinstance(value, Language) else Language[value.upper()]
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
        elif len(args) == 1:
            if isinstance(args[0], LanguageIn):
                nargs = args
            if isinstance(args[0], (set, list, tuple)):
                for v in args[0]:
                    if not isinstance(v, (Language, str)):
                        raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
                    try:
                        nargs = tuple([x if isinstance(x, Language) else Language[x.upper()] for x in args[0]])
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
            else:
                if not isinstance(args[0], (LanguageIn, Language, str)):
                    raise OldapErrorType(f'Value is not an instance of "Language", but "{type(args[0]).__name__}".')
                if isinstance(args[0], LanguageIn):
                    nargs = args
                else:
                    try:
                        nargs = args if isinstance(args[0], Language) else (Language[args[0].upper()],)
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
        else:
            for v in args:
                if not isinstance(v, (Language, str)):
                    raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
                try:
                    nargs = tuple([x if isinstance(x, Language) else Language[x.upper()] for x in args])
                except KeyError as err:
                    raise OldapErrorKey(str(err))

        super().__init__(*nargs, value=nvalue)

    # def __str__(self):
    #     langlist = {f'{x.name.lower()}' for x in self}
    #     return f'({", ".join(langlist)})'
    #
    # def __repr__(self):
    #     langlist = {f'"{x.name.lower()}"' for x in self}
    #     return 'LanguageIn(' + ", ".join(langlist) + ')'

    def __contains__(self, val: Language | str) -> bool:
        if not isinstance(val, Language):
            val = Language[str(val).upper()]
        return super().__contains__(val)

    def add(self, language: Language | Xsd_string | str):
        if not isinstance(language, Language):
            try:
                language = Language[str(language).upper()]
            except ValueError as err:
                raise OldapErrorValue(str(err))
            except KeyError as err:
                raise OldapErrorKey(str(err))
        super().add(language)

    def discard(self, language: Language | Xsd_string | str):
        if not isinstance(language, Language):
            try:
                language = Language[str(language).upper()]
            except ValueError as err:
                raise OldapErrorValue(str(err))
        super().discard(language)




import unittest
from pathlib import Path
from time import sleep

from oldaplib.src.connection import Connection
from oldaplib.src.datamodel import DataModel, PropertyClassChange, ResourceClassChange
from oldaplib.src.dtypes.languagein import LanguageIn
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.enums.language import Language
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.resourceclass import ResourceClass


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestDataModel(unittest.TestCase):
    _context: Context
    _connection: Connection
    _project: Project
    _dmproject: Project
    _sysproject: Project

    @classmethod
    def setUp(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)

        cls._context = Context(name="DEFAULT")
        cls._context['test'] = NamespaceIRI("http://oldap.org/test#")
        cls._context['dmtest'] = NamespaceIRI('http://oldap.org/dmtest#')
        cls._context.use('test', 'dmtest')

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")

        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        cls._connection.clear_graph(Xsd_QName('dmtest:shacl'))
        cls._connection.clear_graph(Xsd_QName('dmtest:onto'))

        file = project_root / 'oldaplib' / 'testdata' / 'connection_test.trig'
        cls._connection.upload_turtle(file)


        sleep(1)  # upload may take a while...

        cls._project = Project.read(cls._connection, "test")
        cls._dmproject = Project.read(cls._connection, "dmtest")
        cls._sysproject = Project.read(cls._connection, "oldap")


    def tearDown(self):
        pass

    def generate_a_datamodel(self, project: Project) -> DataModel:
        dm_name = project.projectShortName
        #
        # define an external standalone property
        #
        comment = PropertyClass(con=self._connection,
                                project=self._dmproject,
                                property_class_iri=Iri(f'{dm_name}:comment'),
                                datatype=XsdDatatypes.langString,
                                name=LangString(["Comment@en", "Kommentar@de"]),
                                uniqueLang=Xsd_boolean(True),
                                languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))
        comment.force_external()

        #
        # Define the properties for the "Book"
        #
        title = PropertyClass(con=self._connection,
                              project=self._dmproject,
                              property_class_iri=Iri(f'{dm_name}:title'),
                              datatype=XsdDatatypes.langString,
                              name=LangString(["Title@en", "Titel@de"]),
                              description=LangString(["Title of book@en", "Titel des Buches@de"]),
                              minCount=Xsd_integer(1),
                              uniqueLang=Xsd_boolean(True),
                              languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                              order=Xsd_decimal(1))

        authors = PropertyClass(con=self._connection,
                                project=self._dmproject,
                                property_class_iri=Iri(f'{dm_name}:authors'),
                                toNodeIri=Iri('oldap:Person'),
                                name=LangString(["Author(s)@en", "Autor(en)@de"]),
                                description=LangString(["Writers of the Book@en", "Schreiber*innen des Buchs@de"]),
                                minCount=Xsd_integer(1),
                                order=Xsd_decimal(2))

        book = ResourceClass(con=self._connection,
                             project=self._dmproject,
                             owlclass_iri=Iri(f'{dm_name}:Book'),
                             label=LangString(["Book@en", "Buch@de"]),
                             comment=LangString("Ein Buch mit Seiten@en"),
                             closed=Xsd_boolean(True),
                             properties=[title, authors, comment])

        pagenum = PropertyClass(con=self._connection,
                                project=self._dmproject,
                                property_class_iri=Iri(f'{dm_name}:pagenum'),
                                datatype=XsdDatatypes.int,
                                name=LangString(["Pagenumber@en", "Seitennummer@de"]),
                                maxCount=Xsd_integer(1),
                                minCount=Xsd_integer(1),
                                order=Xsd_decimal(1))

        inbook = PropertyClass(con=self._connection,
                               project=self._dmproject,
                               property_class_iri=Iri(f'{dm_name}:inbook'),
                               toNodeIri=Iri(f'{dm_name}:Book'),
                               name=LangString(["Pagenumber@en", "Seitennummer@de"]),
                               maxCount=Xsd_integer(1),
                               minCount=Xsd_integer(1),
                               order=Xsd_decimal(1))

        page = ResourceClass(con=self._connection,
                             project=self._dmproject,
                             owlclass_iri=Iri(f'{dm_name}:Page'),
                             label=LangString(["Page@en", "Seite@de"]),
                             comment=LangString("Page of a book@en"),
                             closed=Xsd_boolean(True),
                             properties=[pagenum, inbook, comment])

        dm = DataModel(con=self._connection,
                       project=self._dmproject,
                       propclasses=[comment],
                       resclasses=[book, page])
        return dm

    # @unittest.skip('Work in progress')
    def test_datamodel_constructor(self):
        dm_name = self._dmproject.projectShortName

        dm = self.generate_a_datamodel(self._dmproject)
        dm.create()

        del dm

        dm2 = DataModel.read(con=self._connection, project=self._dmproject)
        p1 = dm2[Iri(f'{dm_name}:comment')]
        self.assertEqual(p1.datatype, XsdDatatypes.langString)
        self.assertEqual(p1.name, LangString(["Comment@en", "Kommentar@de"]))
        self.assertEqual(p1.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))
        self.assertTrue(p1.uniqueLang)

        r1 = dm2[Iri(f'{dm_name}:Book')]

        r1p1 = r1[Iri(f'{dm_name}:title')]
        self.assertEqual(r1p1.internal, Iri(f'{dm_name}:Book'))
        self.assertEqual(r1p1.datatype, XsdDatatypes.langString)
        self.assertEqual(r1p1.name, LangString(["Title@en", "Titel@de"]))
        self.assertEqual(r1p1.description, LangString(["Title of book@en", "Titel des Buches@de"]))
        self.assertEqual(r1p1.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))
        self.assertTrue(r1p1.uniqueLang)
        self.assertEqual(r1p1.minCount, Xsd_integer(1))
        self.assertEqual(r1p1.order, Xsd_decimal(1))

        r1p2 = r1[Iri(f'{dm_name}:authors')]
        self.assertEqual(r1p2.internal, Iri(f'{dm_name}:Book'))
        self.assertEqual(r1p2.toNodeIri, Iri('oldap:Person'))
        self.assertEqual(r1p2.name, LangString(["Author(s)@en", "Autor(en)@de"]))
        self.assertEqual(r1p2.description, LangString(["Writers of the Book@en", "Schreiber*innen des Buchs@de"]))
        self.assertEqual(r1p2.minCount, Xsd_integer(1))
        self.assertEqual(r1p2.order, Xsd_decimal(2))

        r1p3 = r1[Iri(f'{dm_name}:comment')]
        self.assertIsNone(r1p3.internal)
        self.assertEqual(r1p3.datatype, XsdDatatypes.langString)
        self.assertEqual(r1p3.name, LangString(["Comment@en", "Kommentar@de"]))
        self.assertTrue(r1p3.uniqueLang, Xsd_boolean(True))
        self.assertEqual(r1p3.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))

        r2 = dm2[Iri(f'{dm_name}:Page')]
        r2p1 = r2[Iri(f'{dm_name}:pagenum')]
        self.assertEqual(r2p1.internal, Iri(f'{dm_name}:Page'))
        self.assertEqual(r2p1.datatype, XsdDatatypes.int)
        self.assertEqual(r2p1.name, LangString(["Pagenumber@en", "Seitennummer@de"]))
        self.assertEqual(r2p1.maxCount, Xsd_integer(1))
        self.assertEqual(r2p1.minCount, Xsd_integer(1))

        r2p2 = r2[Iri(f'{dm_name}:inbook')]
        self.assertEqual(r2p2.internal, Iri(f'{dm_name}:Page'))
        self.assertEqual(r2p2[PropClassAttr.TO_NODE_IRI], Iri(f'{dm_name}:Book'))
        self.assertEqual(r2p2[PropClassAttr.NAME], LangString(["Pagenumber@en", "Seitennummer@de"]))
        self.assertEqual(r2p2[PropClassAttr.MAX_COUNT], Xsd_integer(1))
        self.assertEqual(r2p2[PropClassAttr.MIN_COUNT], Xsd_integer(1))
        self.assertEqual(r2p2[PropClassAttr.ORDER], Xsd_decimal(1))

        r2p3 = r1[Iri(f'{dm_name}:comment')]
        self.assertIsNone(r2p3.internal)
        self.assertEqual(r2p3.datatype, XsdDatatypes.langString)
        self.assertEqual(r2p3.name, LangString(["Comment@en", "Kommentar@de"]))
        self.assertTrue(r2p3.uniqueLang, Xsd_boolean(True))
        self.assertEqual(r2p3.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))

    # @unittest.skip('Work in progress')
    def test_datamodel_read(self):
        model = DataModel.read(self._connection, self._sysproject)
        print(model.get_propclasses())
        self.assertEqual(set(model.get_propclasses()), {
            Iri("oldap:test"),
            Iri("dcterms:creator"),
            Iri("rdfs:label"),
            Iri("rdfs:comment"),
            Iri("dcterms:created"),
            Iri("dcterms:contributor"),
            Iri("dcterms:modified")
        })
        self.assertTrue(set(model.get_resclasses()) == {
            Iri("oldap:Project"),
            Iri("oldap:User"),
            Iri("oldap:OldapList"),
            Iri("oldap:OldapListNode"),
            Iri("oldap:AdminPermission"),
            Iri("oldap:DataPermission"),
            Iri("oldap:PermissionSet"),
            Iri("oldap:Thing")
        })

    # @unittest.skip('Work in progress')
    def test_datamodel_modify_A(self):
        dm_name = self._dmproject.projectShortName

        dm = self.generate_a_datamodel(self._dmproject)
        dm.create()
        dm = DataModel.read(self._connection, self._dmproject)

        #
        # define an external standalone property
        #
        pubyear = PropertyClass(con=self._connection,
                                project=self._dmproject,
                                property_class_iri=Xsd_QName(f'{dm_name}:pubYear'),
                                datatype=XsdDatatypes.gYear,
                                name=LangString(["Publication Year@en", "Publicationsjahr@de"]),
                                maxCount=Xsd_integer(1))
        pubyear.force_external()
        dm[Iri(f'{dm_name}:pubYear')] = pubyear
        self.assertEqual({Iri(f'{dm_name}:pubYear'): PropertyClassChange(None, Action.CREATE)}, dm.changeset)

        dm[Iri(f'{dm_name}:comment')].name[Language.FR] = 'Commentaire'
        self.assertEqual({
            Iri(f'{dm_name}:pubYear'): PropertyClassChange(None, Action.CREATE),
            Iri(f'{dm_name}:comment'): PropertyClassChange(None, Action.MODIFY)
        }, dm.changeset)

        dm[Iri(f'{dm_name}:Book')][Iri(f'{dm_name}:authors')].name[Language.FR] = "Ecrivain(s)"
        self.assertEqual({
            Iri(f'{dm_name}:pubYear'): PropertyClassChange(None, Action.CREATE),
            Iri(f'{dm_name}:comment'): PropertyClassChange(None, Action.MODIFY),
            Iri(f'{dm_name}:Book'): ResourceClassChange(None, Action.MODIFY)
        }, dm.changeset)

        del dm[Iri(f'{dm_name}:Page')][Iri(f'{dm_name}:comment')]

        self.assertEqual({
            Iri(f'{dm_name}:pubYear'): PropertyClassChange(None, Action.CREATE),
            Iri(f'{dm_name}:comment'): PropertyClassChange(None, Action.MODIFY),
            Iri(f'{dm_name}:Book'): ResourceClassChange(None, Action.MODIFY),
            Iri(f'{dm_name}:Page'): ResourceClassChange(None, Action.MODIFY)
        }, dm.changeset)

        pagename = PropertyClass(con=self._connection,
                                 #graph=dm_name,
                                 project=self._dmproject,
                                 property_class_iri=Xsd_QName(f'{dm_name}:pageName'),
                                 datatype=XsdDatatypes.string,
                                 name=LangString(["Page name@en", "Seitenbezeichnung@de"]),
                                 maxCount=Xsd_integer(1),
                                 minCount=Xsd_integer(1))

        dm[Iri(f'{dm_name}:Page')][Iri(f'{dm_name}:pageName')] = pagename
        self.assertEqual({
            Iri(f'{dm_name}:pubYear'): PropertyClassChange(None, Action.CREATE),
            Iri(f'{dm_name}:comment'): PropertyClassChange(None, Action.MODIFY),
            Iri(f'{dm_name}:Book'): ResourceClassChange(None, Action.MODIFY),
            Iri(f'{dm_name}:Page'): ResourceClassChange(None, Action.MODIFY)
        }, dm.changeset)

    # @unittest.skip('Work in progress')
    def test_datamodel_modify_B(self):
        dm_name = self._dmproject.projectShortName

        dm = self.generate_a_datamodel(self._dmproject)
        dm.create()
        del dm

        dm = DataModel.read(self._connection, self._dmproject)

        #
        # define an external standalone property
        #
        pubyear = PropertyClass(con=self._connection,
                                project=self._dmproject,
                                property_class_iri=Xsd_QName(f'{dm_name}:pubYear'),
                                datatype=XsdDatatypes.gYear,
                                name=LangString(["Publication Year@en", "Publicationsjahr@de"]),
                                maxCount=Xsd_integer(1))
        pubyear.force_external()

        dm[Iri(f'{dm_name}:pubYear')] = pubyear
        dm[Iri(f'{dm_name}:comment')][PropClassAttr.NAME][Language.FR] = 'Commentaire'
        dm[Iri(f'{dm_name}:Book')][Iri(f'{dm_name}:authors')].name[Language.FR] = "Ecrivain(s)"
        del dm[Iri(f'{dm_name}:Page')][Iri(f'{dm_name}:comment')]

        pagename = PropertyClass(con=self._connection,
                                 #graph=dm_name,
                                 project=self._dmproject,
                                 property_class_iri=Xsd_QName(f'{dm_name}:pageName'),
                                 datatype=XsdDatatypes.string,
                                 name=LangString(["Page name@en", "Seitenbezeichnung@de"]),
                                 maxCount=Xsd_integer(1),
                                 minCount=Xsd_integer(1))

        dm[Iri(f'{dm_name}:Page')][Iri(f'{dm_name}:pageName')] = pagename

        dm.update()

        del dm

        dm = DataModel.read(self._connection, self._dmproject)
        self.assertIsNotNone(dm.get(Iri(f'{dm_name}:pubYear')))
        self.assertEqual(dm[Iri(f'{dm_name}:pubYear')].datatype, XsdDatatypes.gYear)
        self.assertEqual(dm[Iri(f'{dm_name}:pubYear')].maxCount, 1)
        self.assertEqual(dm[Iri(f'{dm_name}:comment')].name[Language.FR], 'Commentaire')
        self.assertEqual(dm[Iri(f'{dm_name}:Book')][Iri(f'{dm_name}:authors')].name[Language.FR], "Ecrivain(s)")
        self.assertIsNotNone(dm[Iri(f'{dm_name}:Page')][Iri(f'{dm_name}:pageName')])
        self.assertIsNone(dm[Iri(f'{dm_name}:Page')].get(Iri(f'{dm_name}:comment')))

    def test_write_trig(self):
        pagename = PropertyClass(con=self._connection,
                                 project=self._project,
                                 property_class_iri=Iri('test:pageDesignation'),
                                 datatype=XsdDatatypes.string,
                                 name=LangString("Page designation@en", "Seitenbezeichnung@de"),
                                 minCount=Xsd_integer(1),
                                 order=Xsd_decimal(1))
        pagenum = PropertyClass(con=self._connection,
                                project=self._project,
                                property_class_iri=Iri('test:pageNum'),
                                datatype=XsdDatatypes.positiveInteger,
                                name=LangString("Pagenumber@en", "Seitennummer@de"),
                                description=LangString("consecutive numbering of pages@en", "Konsekutive Nummerierung der Seiten@de"),
                                minCount=Xsd_integer(1),
                                maxCount=Xsd_integer(1),
                                order=Xsd_decimal(2))
        pagedescription = PropertyClass(con=self._connection,
                                        project=self._project,
                                        property_class_iri=Iri('test:pageDescription'),
                                        datatype=XsdDatatypes.langString,
                                        languageIn=LanguageIn(Language.EN, Language.DE),
                                        uniqueLang=Xsd_boolean(True),
                                        order=Xsd_decimal(3))
        content = PropertyClass(con=self._connection,
                                project=self._project,
                                property_class_iri=Iri('test:pageContent'),
                                datatype=XsdDatatypes.string,
                                maxCount=Xsd_integer(1),
                                order=Xsd_decimal(4))
        inBook = PropertyClass(con=self._connection,
                               project=self._project,
                               property_class_iri=Iri('test:pageInBook'),
                               toNodeIri=Iri('test:Book'),
                               minCount=Xsd_integer(1),
                               maxCount=Xsd_integer(1),
                               order=Xsd_decimal(5))
        page = ResourceClass(con=self._connection,
                             project=self._project,
                             owlclass_iri=Iri("test:Page"),
                             superclass=Iri('oldap:Thing'),
                             label=LangString(["Project@en", "Projekt@de"]),
                             comment=LangString(["A page of a book@en", "Seite eines Buches@de"]),
                             closed=Xsd_boolean(True),
                             properties=[pagename, pagenum, pagedescription, content, inBook])

        title = PropertyClass(con=self._connection,
                              project=self._project,
                              property_class_iri=Iri('test:title'),
                              datatype=XsdDatatypes.string,
                              minCount=Xsd_integer(1),
                              maxCount=Xsd_integer(1),
                              order=Xsd_decimal(1))
        author = PropertyClass(con=self._connection,
                               project=self._project,
                               property_class_iri=Iri('test:author'),
                               toNodeIri=Iri('test:Person'),
                               order=Xsd_decimal(2))
        pubDate = PropertyClass(con=self._connection,
                                project=self._project,
                                property_class_iri=Iri('test:pubDate'),
                                datatype=XsdDatatypes.date,
                                minCount=Xsd_integer(1),
                                maxCount=Xsd_integer(1),
                                order=Xsd_decimal(3))
        book = ResourceClass(con=self._connection,
                             project=self._project,
                             owlclass_iri=Iri('test:Book'),
                             superclass=Iri('oldap:Thing'),
                             label=LangString(["Book@en", "Buch@de"]),
                             closed=Xsd_boolean(True),
                             properties=[title, author, pubDate])

        person = ResourceClass(con=self._connection,
                               project=self._project,
                               owlclass_iri=Iri('test:Person'),
                               superclass=Iri('oldap:Thing'),
                               label=LangString(["Person@en", "Person@de"]),
                               properties=[title, author, pubDate])

        dm = DataModel(con=self._connection,
                       project=self._project,
                       resclasses=[page, book])
        dm.write_as_trig('gaga.trig')





"""
This module provides the handling for language dependent strings. RDF allows to attach a language tag to
a string to identify which language is used. In turtle/trig notation it has the form "string"@lang,
eg "this is a string in english"@en. In order to handle such strings easily in OMAS,
the Language class provides all necessary enumerations.
"""
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, Callable, Self

from pystrict import strict

from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.enums.action import Action
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorValue
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd_string import Xsd_string


@dataclass
class LangStringChange:
    old_value: str | None
    action: Action

#@strict
@serializer
class LangString(Notify):
    """
    Implements a multi-language representation of a string.

    RDF allows to tag strings with language identifieres, e.g. "this is a text"@en . To support multiple languages it
    may make sense to allow for some string properties a cardinality > 1 so that multiple strings in different labguages
    may be assigned. The SHACL restriction sh:uniqueLang allows to restrict the string to a unique value for each
    language. This class uses a Dict with the language as key (see Language.py) and the string as value. For strings
    with no attached language the Language.XX tag is used. Methods that require a language parameter, it can usually
    be given as member of the enum class *Language*, e.g. ``Language.EN`` as 2-character shortname, e.g. ``"en"``. The
    use of the enumeration is strongly recommended.
    An instance of this class implements some dictionary behaviours:

    - access to a specific language string representation: `myvar[Language.FR]` or `myvar["fr"]`
    - to add or replace a specific language string representation: `myvar[Language.DE] = "Ein neuer string"` or
      `myvar["de"] = "Ein neuer string"`.
    - to delete a language: `del myvar[Language.FR]`

    The class implements the following methods:

    - _LangString()_: Initiate a language string instance (`__init__()`)
    - _len()_: (see `__len__()_`) Return the number of language strings present in the instance
    - _str()_: (see `__str__()`) Return the string representation of the Langstring as it would be used in a SPARQL insert statement
    - _get()_: (see `__get__()`)Return the language string or None, if it does not exist
    - _==_: (see `__eq__()`) Test for equality of 2 langstrings
    - _!=_: (see `__ne__()`) Test for inequality of 2 langstrings
    - _items()_: (see `__items__()')Retuns an iterator over all strings in the LangString instance
    - _langstring_: Returns dict containing all languages present in the instance
    - _add()_: Add a string
    - _undo_(): Forget all changes
    - _changeset()_: Return the changeset dict (Note: this is not for generic use)
    - _changeset_clear()_: Clear changeset to an empty dict
    - _update_shacl_(): Return the SPARQL code piece that updates a Language string SHACL part of the triple store.
    - _delete_shacl_(): Return the SPARQL code piece that deletes an LanguageString
    """
    _langstring: Dict[Language, str]
    _changeset: Dict[Language, LangStringChange]
    _notifier: Callable[[type], None] | None

    defaultLanguage: Language = Language.EN
    priorities: list[Language] = [Language.EN, Language.DE, Language.FR]

    def __init__(self, *args: str | Xsd_string | List[str] | Dict[Language | str, str] | Self | None,
                 langstring: str | Xsd_string | List[str] | Dict[Language | str, str] | Self | None = None,
                 notifier: Optional[Callable[[PropClassAttr], None]] = None,
                 notify_data: Optional[PropClassAttr] = None):
        """
        Implements language dependent strings.

        The parameter `langstring`can either be

        - a string in the form `"string@ll"`, eg `"Lastname@en"`. A string without language qualifier has the
            language `Language.XX` associated.
        - A list of strings: `["Lastname@en", "Nachname@de"]`
        - A dict with language short names as key: `{'en': "Lastname", 'de': "Nachname"}`
        - A dict with Language enum values as keys: `{Language.EN: "Lastname, Language.DE: "Nachname"}`

        :param langstring: A definition of one or several langiage dependent strings.
        :param priorities: If a desired language is not found, then the next string is used given this priority list
          which as the form [Langguage.LL, ...], eg [Language.EN, Language.DE, Language.XX]. The default value
          is [Language.XX, Language.EN, Language.DE, Language.FR]
        """
        super().__init__(notifier, notify_data)
        self._changeset = {}
        self._langstring = {}

        if len(args) <= 1:
            if len(args) == 1:
                langstring = args[0]
            if langstring is None:
                return
            else:
                if isinstance(langstring, LangString):
                    self._langstring = langstring._langstring
                elif isinstance(langstring, Xsd_string):
                    if not langstring:
                        return
                    l = LangString.defaultLanguage if langstring.lang is None else langstring.lang
                    self._langstring[l] = langstring.value
                elif isinstance(langstring, str):
                    if not langstring:
                        return
                    xstr = Xsd_string(langstring)
                    l = LangString.defaultLanguage if xstr.lang is None else xstr.lang
                    self._langstring[l] = xstr.value
                elif isinstance(langstring, (list, tuple)):
                    for lstr in langstring:
                        xstr = Xsd_string(lstr)
                        if not xstr:
                            continue
                        l = LangString.defaultLanguage if xstr.lang is None else xstr.lang
                        self._langstring[l] = xstr.value
                elif isinstance(langstring, dict):
                    for lang, value in langstring.items():
                        xstr = Xsd_string(value, lang)
                        if not xstr:
                            continue
                        l = LangString.defaultLanguage if xstr.lang is None else xstr.lang
                        self._langstring[l] = xstr.value
                else:
                    raise OldapErrorValue(f'LangString parameter has wrong datatype: {type(langstring).__name__}, must be "str | Xsd_string | List[str] | Dict[Language | str, str] | LangString"')
        else:
            for langstring in args:
                if isinstance(langstring, Xsd_string):
                    if langstring:
                        l = LangString.defaultLanguage if langstring.lang is None else langstring.lang
                        self._langstring[l] = langstring.value
                elif isinstance(langstring, str):
                    xstr = Xsd_string(langstring)
                    if not xstr:
                        continue
                    l = LangString.defaultLanguage if xstr.lang is None else xstr.lang
                    self._langstring[l] = xstr.value
                else:
                    raise OldapErrorValue(
                        f'LangString parameter has wrong datatype: {type(langstring).__name__}, must be "str | Xsd_string | List[str] | Dict[Language | str, str] | LangString"')

    def __len__(self):
        """
        Returns the number of languages defined for the given the LangString instance
        :return: Number of languages defined
        """
        return len(self._langstring)

    def __bool__(self):
        return len(self._langstring) > 0

    def __getitem__(self, lang: str | Language) -> str:
        """
        Get the string of the given language.
        :param lang: The desired language, either as string shortname or as Language enum
        :return: The string – may return '--no string--' as placeholder of language is not available
        """
        if isinstance(lang, str):
            try:
                lang = Language[lang.upper()]
            except KeyError:
                raise OldapError(f'Language "{lang}" is invalid')
        s = self._langstring.get(lang)
        if s:
            return s
        else:
            for ll in self.priorities:
                if self._langstring.get(ll) is not None:
                    return self._langstring[ll]
            return '--no string--'

    def __setitem__(self, lang: Language | str, value: str) -> None:
        """
        Set a new or change an existing language steing
        :param lang: Language as shortname or Language enum
        :param value: The string value
        :return: None
        """
        if isinstance(lang, Language):
            if self._changeset.get(lang) is None:  # only the first change is recorded
                self._changeset[lang] = LangStringChange(self._langstring.get(lang),
                                                         Action.REPLACE if self._langstring.get(lang) else Action.CREATE)
            self._langstring[lang] = value
            self.notify()
        elif isinstance(lang, str):
            try:
                lobj = Language[lang.upper()]
                if self._changeset.get(lobj) is None:  # only the first change is recorded
                    self._changeset[lobj] = LangStringChange(self._langstring.get(lobj),
                                                             Action.REPLACE if self._langstring.get(lobj) else Action.CREATE)
                self._langstring[lobj] = value
                self.notify()
            except (KeyError, ValueError) as err:
                raise OldapError(f'Language "{lang}" is invalid: {err}.')
        else:
            raise OldapError(f'Language "{lang}" is invalid.')

    def __delitem__(self, lang: Language | str) -> None:
        """
        Delete a given language from a language string
         :param lang: The language (as short name of as Language enum) to be deleted
        :type lang: Language or str (shortname)
        :return: Does return nothing
        :rtype: None
        """
        if isinstance(lang, Language):
            try:
                if self._changeset.get(lang) is None:
                    self._changeset[lang] = LangStringChange(self._langstring[lang], Action.DELETE)
                del self._langstring[lang]
                self.notify()
            except KeyError as err:
                raise OldapError(f'No language string of language: "{lang}"!')
        elif isinstance(lang, str):
            try:
                lobj = Language[lang.upper()]
                if self._changeset.get(lobj) is None:
                    self._changeset[lobj] = LangStringChange(self._langstring[lobj], Action.DELETE)
                del self._langstring[lobj]
                self.notify()
            except (KeyError, ValueError) as err:
                raise OldapError(f'No language string of language: "{lang}"!')
        else:
            raise OldapError(f'Unsupported language value {lang}!')

    def __str__(self) -> str:
        """
        Return the language string as it would be used in a SPARQL insert statement
        :return: language string as it would be used in a SPARQL insert statement
        :rtype: str
        """
        langlist = [f'"{val}@{lang.name.lower()}"' for lang, val in self._langstring.items()]
        resstr = ", ".join(langlist)
        return resstr

    def __repr__(self) -> str:
        return f'LangString({self.__str__()})'

    @property
    def toRdf(self) -> str:
        langStringList = [Xsd_string(val, lang).toRdf for lang, val in self._langstring.items()]
        resstr = ", ".join(langStringList)
        return resstr

    def _as_dict(self) -> dict:
        """
        Return a dictionary used for JSONification of a LangString instance
        :return: Dict that can be serialized
        """
        return {
            'langstring': [f'"{val}@{lang.value.lower()}"' for lang, val in self._langstring.items()],
            'priorities': [lang.value.lower() for lang in self.priorities]
        }

    def get(self, lang: str | Language, default: str = None) -> str:
        """
        Return the language string or None, if it does not exist
        :param lang: Desired language
        :type lang: Either a string (shortname) or a Language enum element.
        :return: language string
        :rtype: str
        """
        if isinstance(lang, str):
            lang = Language[lang.upper()]
        return self._langstring.get(lang, default)

    def __bool__(self) -> bool:
        return len(self._langstring) > 0

    def __eq__(self, other: Self | None) -> bool:
        """
        Test for equality of two language strings
        :param other: The other Language string to compare to
        :type other: LanguageString
        :return: True or False
        :rtype: bool
        """
        if other is None:
            return False
        if len(self._langstring) != len(other._langstring):
            return False
        for lang in self._langstring:
            if other.langstring.get(lang) is None:
                return False
            if self._langstring.get(lang) != other.langstring.get(lang):
                return False
        return True

    def __ne__(self, other: Self) -> bool:
        """
        Test for inequality of two language strings
        :param other: The other Language string to compare to
        :type other: LanguageString
        :return: True if inequal, otherweise False
        :rtype: bool
        """
        if len(self._langstring) != len(other._langstring):
            return True
        for lang in self._langstring:
            if other.langstring.get(lang) is None:
                return True
            if self._langstring.get(lang) != other.langstring.get(lang):
                return True
        return False

    def items(self):
        """
        Return an iterator over the language strings
        :return: iterator over the language strings
        :rtype: iterator
        """
        return self._langstring.items()

    @property
    def langstring(self) -> Dict[Language, str]:
        """
        All language strings as Dict
        :return: A dictionary of all language strings
        :rtype: Dict
        """
        return self._langstring

    def add(self, *args: str | Xsd_string | List[str] | Dict[Language | str, str] | Self) -> None:
        """
        Add one or several new languages to a lang string. The method accepts several forms:
        * ``mylstr.add("a new string@en")``
        * ``mylstr.add(["a new string@en", "eine neue Zeichenketter@de])``
        * ``mylstr.add({"fr": "Nouveau", "de": "Neue Zeichenketter"})``
        * ``mylstr.add({Language.FR: "Nouveau", Language.DE: "Neue Zeichenketter"})
        As this, it's a very versatile method for adding new languages to a LanguageString instance
        :param langs: The language/string pairs as single value, list or dict
        :type langs: str | List[str] | Dict[str, str] | Dict[Language, str]
        :return: No return value
        :rtype: None
        """
        if len(args) == 0:
            return
        elif len(args) == 1:
            if isinstance(args[0], LangString):
                for lang, val in args[0].langstring.items():
                    oldval = self._langstring.get(lang)
                    self._langstring[lang] = val
                    if self._changeset.get(lang) is None:  # only the first change is recorded
                        self._changeset[lang] = LangStringChange(oldval,
                                                                 Action.REPLACE if oldval is not None else Action.CREATE)
            elif isinstance(args[0], Xsd_string):
                l = LangString.defaultLanguage if args[0].lang is None else args[0].lang
                oldval = self._langstring.get(l)
                self._langstring[l] = args[0].value
                if self._changeset.get(l) is None:  # only the first change is recorded
                    self._changeset[l] = LangStringChange(oldval,
                                                             Action.REPLACE if oldval is not None else Action.CREATE)
            elif isinstance(args[0], str):
                xstr = Xsd_string(args[0])
                l = xstr.lang or LangString.defaultLanguage
                oldval = self._langstring.get(l)
                self._langstring[l] = xstr.value
                if self._changeset.get(l) is None:  # only the first change is recorded
                    self._changeset[l] = LangStringChange(oldval, Action.REPLACE if oldval is not None else Action.CREATE)
            elif isinstance(args[0], (list, tuple)):
                for lstr in args[0]:
                    xstr = Xsd_string(lstr)
                    l = xstr.lang or LangString.defaultLanguage
                    oldval = self._langstring.get(l)
                    self._langstring[l] = xstr.value
                    if self._changeset.get(l) is None:  # only the first change is recorded
                        self._changeset[l] = LangStringChange(oldval, Action.REPLACE if oldval is not None else Action.CREATE)
            elif isinstance(args[0], dict):
                for lang, value in args[0].items():
                    xstr = Xsd_string(value, lang)
                    l = xstr.lang or LangString.defaultLanguage
                    oldval = self._langstring.get(l)
                    self._langstring[l] = xstr.value
                    if self._changeset.get(l) is None:  # only the first change is recorded
                        self._changeset[l] = LangStringChange(oldval,
                                                              Action.REPLACE if oldval is not None else Action.CREATE)
            else:
                raise OldapErrorValue(
                    f'LangString parameter has wrong datatype: {type(args[0]).__name__}, must be "str | Xsd_string | List[str] | Dict[Language | str, str] | LangString"')
        else:
            for langstring in args:
                if isinstance(langstring, Xsd_string):
                    l = langstring.lang or LangString.defaultLanguage
                    oldval = self._langstring.get(l)
                    self._langstring[l] = langstring.value
                    if self._changeset.get(l) is None:  # only the first change is recorded
                        self._changeset[l] = LangStringChange(oldval,
                                                              Action.REPLACE if oldval is not None else Action.CREATE)
                elif isinstance(langstring, str):
                    xstr = Xsd_string(langstring)
                    l = xstr.lang or LangString.defaultLanguage
                    oldval = self._langstring.get(l)
                    self._langstring[l] = xstr.value
                    if self._changeset.get(l) is None:  # only the first change is recorded
                        self._changeset[l] = LangStringChange(oldval,
                                                              Action.REPLACE if oldval is not None else Action.CREATE)
                else:
                    raise OldapErrorValue(
                        f'LangString parameter has wrong datatype: {type(langstring).__name__}, must be "str | Xsd_string | List[str] | Dict[Language | str, str] | LangString"')

    def undo(self) -> None:
        """
        Undo all changes made since last update/creation/read
        :return: Nothing
        :rtype: None
        """
        for lang, change in self._changeset.items():
            if change.action == Action.CREATE:
                del self._langstring[lang]
            else:
                self._langstring[lang] = change.old_value
        self._changeset = {}

    @property
    def changeset(self) -> Dict[Language, LangStringChange]:
        """
        Return the changeset dict (Note: this is not for generic use)
        :return: The changeset information
        :rtype: Dict[Language, LangStringChange]
        """
        return self._changeset

    def changeset_clear(self) -> None:
        """
        Clear changeset to an empty dict
        :return: Nothing
        :rtype: None
        """
        self._changeset = {}

    def create(self, *,
               graph: Xsd_QName,
               subject: Iri,
               field: Xsd_QName,
               indent: int = 0, indent_inc: int = 4):
        blank = ''
        sparql_list = []
        sparql = ''
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {graph} {{\n'
        sparql += f'{blank:{(indent + 2) * indent_inc}}{subject.toRdf} {field.toRdf} {self.toRdf} .\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'
        return sparql

    def delete(self, *,
               graph: Xsd_QName,
               subject: Iri,
               field: Xsd_QName,
               indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql_list = []
        sparql = ''
        sparql += f'{blank:{indent * indent_inc}}DELETE WHERE {{\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {graph} {{\n'
        sparql += f'{blank:{(indent + 2) * indent_inc}}{subject.toRdf} {field.toRdf} ?o .\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'
        return sparql

    def update(self, *,
               graph: Xsd_QName,
               subject: Iri,
               subjectvar: str,
               field: Xsd_QName,
               indent: int = 0, indent_inc: int = 4) -> List[str]:
        blank = ''
        sparql_list = []
        for lang, change in self._changeset.items():
            if change.action != Action.CREATE:
                sparql = f'{blank:{indent * indent_inc}}DELETE DATA {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {graph} {{\n'
                tmpstr = f'"{change.old_value}"'
                tmpstr += "@" + lang.name.lower()
                sparql += f'{blank:{(indent + 2) * indent_inc}}{subject.toRdf} {field.toRdf} {tmpstr} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)
            if change.action != Action.DELETE:
                sparql = f'{blank:{indent * indent_inc}}INSERT DATA {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {graph} {{\n'
                langstr = f'"{self._langstring[lang]}"'
                langstr += "@" + lang.name.lower()
                sparql += f'{blank:{(indent + 2) * indent_inc}}{subject.toRdf} {field.toRdf} {langstr} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)

            # sparql = ''
            # sparql += f'{blank:{indent * indent_inc}}WITH {graph}\n'
            # if change.action != Action.CREATE:
            #     sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
            #     tmpstr = f'"{change.old_value}"'
            #     if lang != Language.XX:
            #         tmpstr += "@" + lang.name.lower()
            #     sparql += f'{blank:{(indent + 1) * indent_inc}}{subjectvar} {repr(field)} {tmpstr} .\n'
            #     sparql += f'{blank:{indent * indent_inc}}}}\n'
            #
            # if change.action != Action.DELETE:
            #     sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
            #     langstr = f'"{self._langstring[lang]}"'
            #     if lang != Language.XX:
            #         langstr += "@" + lang.name.lower()
            #     sparql += f'{blank:{(indent + 1) * indent_inc}}{subjectvar} {repr(field)} {langstr} .\n'
            #     sparql += f'{blank:{indent * indent_inc}}}}\n'
            #
            # sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            # sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({repr(subject)} as {subjectvar}) .\n'
            # if change.action != Action.CREATE:
            #     tmpstr = f'"{change.old_value}"'
            #     if lang != Language.XX:
            #         tmpstr += "@" + lang.name.lower()
            #     sparql += f'{blank:{(indent + 1) * indent_inc}}{subjectvar} {repr(field)} {tmpstr} .\n'
            # sparql += f'{blank:{indent * indent_inc}}}}'
            # sparql_list.append(sparql)
        return sparql_list

    def update_shacl(self, *,
                     graph: Xsd_NCName,
                     owlclass_iri: Iri | None = None,
                     prop_iri: Iri,
                     attr: PropClassAttr,
                     modified: Xsd_dateTime,
                     indent: int = 0, indent_inc: int = 4) -> str:
        """
        Return the SPARQL code piece that updates a Language string SHACL part of the triple store.
        :param graph: SPARQL graph as described in the introduction to OMASLIB
        :type graph: Xsd_NCName
        :param owlclass_iri: The OWL class IRI of the associated ResourceClass. May be omitted for standalone properties
        :type owlclass_iri: Xsd_QName | None
        :param prop_iri: The property IRI of the associated PropertyClass
        :type prop_iri: Xsd_QName
        :param attr: The QName of the associated attribute
        :type attr: Xsd_QName
        :param modified: timestamp that should be applied
        :type modified: datetime
        :param indent: The indent for the generated SPARQL code
        :type indent: int
        :param indent_inc: The indent increment for the generated SPARQL code
        :type indent_inc: int
        :return: SPARQL code piece
        :rtype: str
        """

        blank = ''
        sparql_list = []
        for lang, change in self._changeset.items():
            sparql = f'# LangString: Process "{lang.name}" with Action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH {graph}:shacl\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                tmpstr = f'"{change.old_value}"'
                tmpstr += "@" + lang.name.lower()
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {tmpstr} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                langstr = f'"{self._langstring[lang]}"'
                langstr += "@" + lang.name.lower()
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {langstr} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            if owlclass_iri:
                sparql += f'{blank:{(indent + 1) * indent_inc}}{owlclass_iri}Shape sh:property ?prop .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:path {prop_iri.toRdf} .\n'
            else:
                sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({prop_iri}Shape as ?prop) .\n'
            if change.action != Action.CREATE:
                tmpstr = f'"{change.old_value}"'
                tmpstr += "@" + lang.name.lower()
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {tmpstr} .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop dcterms:modified ?modified .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}FILTER(?modified = {modified.toRdf})\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = ";\n".join(sparql_list)
        return sparql

    def delete_shacl(self, *,
                     graph: Xsd_NCName,
                     owlclass_iri: Iri | None = None,
                     prop_iri: Iri,
                     attr: PropClassAttr,
                     modified: datetime,
                     indent: int = 0, indent_inc: int = 4) -> str:
        # TODO: Include into unit tests!
        """
        Return the SPARQL code piece that deletes an LanguageString
        :param graph: SPARQL graph as described in the introduction to OMASLIB
        :type graph: Xsd_NCName
        :param owlclass_iri: The OWL class IRI of the associated ResourceClass. May be omitted for standalone properties
        :type owlclass_iri: NCName or None
        :param prop_iri: The property IRI of the associated PropertyClass
        :type prop_iri: Xsd_QName
        :param attr: The QName of the associated attribute
        :type attr: Xsd_QName
        :param modified: Modification date to apply
        :type modified: datetime
        :param indent: The indent for the generated SPARQL code
        :type indent: int
        :param indent_inc: The indent increment for the generated SPARQL code
        :type indent_inc: int
        :return: Piece of SPARQL code that deletes the Language String from the SHACL definition
        :rtype: str
        """
        blank = ''
        sparql = f'#\n# Deleting the complete LangString data for {prop_iri} {attr.value}\n#\n'
        sparql += f'{blank:{indent * indent_inc}}WITH {graph}:shacl'
        sparql += f'{blank:{indent * indent_inc}}DELETE {{'
        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} ?langval'
        sparql += f'{blank:{indent * indent_inc}}}}'
        sparql += f'{blank:{indent * indent_inc}}WHERE {{'
        if owlclass_iri:
            sparql += f'{blank:{(indent + 1) * indent_inc}}{owlclass_iri}Shape sh:property ?prop .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:path {prop_iri.toRdf} .\n'
        else:
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({prop_iri}Shape as ?prop)\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} ?langval'
        sparql += f'{blank:{indent * indent_inc}}}}'
        return sparql


if __name__ == '__main__':
    l0 = LangString()
    print("l0: ", l0, len(l0))
    l1 = LangString([])
    print("l1 : ", l1, len(l1))
    l2 = LangString([""])
    print("l2 : ", l2, len(l2))




from abc import ABC, abstractmethod
from typing import Self, Dict


class Xsd(ABC):
    """
    Abstract base class for XSD classes.
    """

    @abstractmethod
    def __init__(self, value: Self | str, validate: bool = True):
        """
        Initialize the XSD class. Must not be called by subclasses
        :param value:
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """
        Return the string representation of the XSD class.
        :return: string representation of the XSD class.
        """
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """
        Return the string representation of the XSD class as it would be for constructing the XSD class.
        :return: string representation of the XSD class.
        """
        pass

    @classmethod
    def fromRdf(cls, value: str) -> Self:
        """
        Create an XSD instance from a RDF value string.
        :param value: String with RDF value string.
        :return: Instance of the XSD class.
        """
        return cls(value, validate=False)

    @property
    @abstractmethod
    def toRdf(self) -> str:
        """
        Property returning the RDF representation of the XSD class.
        :return: String for the RDF representation of the XSD class.
        """
        pass

    @abstractmethod
    def _as_dict(self) -> dict[str, str]:
        """
        Used for the JSON serialization of the XSD class.
        :return:
        """
        pass



import unittest
from pathlib import Path
from time import sleep

from oldaplib.src.ObjectFactory import ResourceInstanceFactory
from oldaplib.src.connection import Connection
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_qname import Xsd_QName


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path

class TestObjectFactory(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)
        cls._context = Context(name="DEFAULT")

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="fornaro",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")
        cls._context['test'] = 'http://oldap.org/test#'
        cls._context.use('test')

        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        cls._connection.clear_graph(Xsd_QName('oldap:admin'))

        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

        file = project_root / 'oldaplib' / 'testdata' / 'objectfactory_test.trig'
        cls._connection.upload_turtle(file)

        sleep(1)  # upload may take a while...

    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(Xsd_QName('oldaplib:admin'))
        #cls._connection.upload_turtle("oldaplib/ontologies/admin.trig")
        #sleep(1)  # upload may take a while...
        pass


    def test_constructor(self):
        project = Project.read(con=self._connection, projectIri_SName='test')
        factory = ResourceInstanceFactory(con=self._connection, project=project)
        Book = factory.createObjectInstance(Iri('test:Book'), 'Book')
        b = Book(title="Hitchhiker's Guide to the Galaxy",
                 author=Iri('test:DouglasAdams'),
                 pubDate="1995-09-27")
        self.assertEqual(b.title, "Hitchhiker's Guide to the Galaxy")
        self.assertEqual(b.author, Iri('test:DouglasAdams'))
        self.assertEqual(b.pubDate, "1995-09-27")

        Page = factory.createObjectInstance(Iri('test:Page'), 'Page')
        p1 = Page(pageDesignation="Cover",
                 pageNum=1,
                 pageDescription=LangString("Cover page of book@en", "Vorderseite des Bucheinschlags@de"),
                 pageInBook="test:Hitchhiker")
        self.assertEqual(p1.pageDesignation, "Cover")
        self.assertEqual(p1.pageNum, 1)
        self.assertEqual(p1.pageDescription, LangString("Cover page of book@en", "Vorderseite des Bucheinschlags@de"))
        self.assertEqual(p1.pageInBook, "test:Hitchhiker")

if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.oldaperror import OldapErrorInconsistency, OldapError, OldapErrorValue
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.semantic_version import SemanticVersion
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.resourceclass import ResourceClass

@dataclass
class ResourceClassChange:
    old_value: ResourceClass | None
    action: Action

@dataclass
class PropertyClassChange:
    old_value: PropertyClass | None
    action: Action

class DataModel(Model):
    __graph: Xsd_NCName
    _project: Project
    __context: Context
    __version: SemanticVersion
    __propclasses: dict[Iri, PropertyClass | None]
    __resclasses: dict[Iri, ResourceClass | None]
    __resclasses_changeset: dict[Iri, ResourceClassChange]
    __propclasses_changeset: dict[Iri, PropertyClassChange]

    def __init__(self, *,
                 con: IConnection,
                 project: Project,
                 propclasses: list[PropertyClass] | None = None,
                 resclasses: list[ResourceClass] | None = None) -> None:
        super().__init__(con)
        self.__version = SemanticVersion()

        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        self._project = project
        context = Context(name=self._con.context_name)
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        self.__graph = project.projectShortName

        self.__propclasses = {}
        if propclasses is not None:
            for p in propclasses:
                self.__propclasses[p.property_class_iri] = p
        self.__resclasses = {}
        if resclasses is not None:
            for r in resclasses:
                self.__resclasses[r.owl_class_iri] = r
        self.__propclasses_changeset = {}
        self.__resclasses_changeset = {}

    def __getitem__(self, key: Iri) -> PropertyClass | ResourceClass:
        if key in self.__resclasses:
            return self.__resclasses[key]
        if key in self.__propclasses:
            return self.__propclasses[key]
        else:
            raise KeyError(key)

    def __setitem__(self, key: Iri, value: PropertyClass | ResourceClass) -> None:
        if isinstance(value, PropertyClass):
            if self.__propclasses.get(key) is None:
                self.__propclasses_changeset[key] = PropertyClassChange(None, Action.CREATE)
            else:
                old_value = deepcopy(self.__propclasses[key])
                self.__propclasses_changeset[key] = PropertyClassChange(old_value, Action.MODIFY)
            self.__propclasses[key] = value
        elif isinstance(value, ResourceClass):
            if self.__resclasses.get(key) is None:
                self.__resclasses_changeset[key] = ResourceClassChange(None, Action.CREATE)
            else:
                old_value = deepcopy(self.__resclasses[key])
                self.__resclasses_changeset[key] = ResourceClassChange(old_value, Action.MODIFY)
            self.__resclasses[key] = value
        else:
            raise OldapErrorValue(f'"{key}" must be either PropertyClass or ResourceClass (is "{type(value).__name__}")')

    def __delitem__(self, key: Iri) -> None:
        if key in self.__propclasses:
            self.__propclasses_changeset[key] = PropertyClassChange(self.__propclasses[key], Action.DELETE)
            del self.__propclasses[key]
        elif key in self.__resclasses:
            self.__resclasses_changeset[key] = ResourceClassChange(self.__resclasses[key], Action.DELETE)
            del self.__resclasses[key]
        else:
            raise OldapErrorValue(f'"{key}" must be either PropertyClass or ResourceClass')

    def get(self, key: Iri) -> PropertyClass | ResourceClass | None:
        if key in self.__propclasses:
            return self.__propclasses[key]
        elif key in self.__resclasses:
            return self.__resclasses[key]
        else:
            return None


    def get_propclasses(self) -> list[Iri]:
        return [x for x in self.__propclasses]

    def get_resclasses(self) -> list[Iri]:
        return [x for x in self.__resclasses]

    @property
    def changeset(self) -> dict[Iri, PropertyClassChange | ResourceClassChange]:
        return self.__resclasses_changeset | self.__propclasses_changeset

    def changeset_clear(self) -> None:
        for prop, change in self.__propclasses_changeset.items():
            if change.action == Action.MODIFY:
                self.__propclasses[prop].changeset_clear()
        self.__propclasses_changeset = {}
        for res, change in self.__resclasses_changeset.items():
            if change.action == Action.MODIFY:
                self.__resclasses[res].changeset_clear()
        self.__resclasses_changeset = {}

    def notifier(self, what: Iri) -> None:
        if what in self.__propclasses:
            self.__propclasses_changeset[what] = PropertyClassChange(None, Action.MODIFY)
        elif what in self.__resclasses:
            self.__resclasses_changeset[what] = ResourceClassChange(None, Action.MODIFY)
        else:
            raise OldapErrorInconsistency(f'No resclass or property "{what}" in datamodel.')

    @classmethod
    def read(cls, con: IConnection,
             project: Project):
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        cls.__context = Context(name=con.context_name)
        cls.__context[project.projectShortName] = project.namespaceIri
        cls.__context.use(project.projectShortName)
        cls.__graph = project.projectShortName
        #
        # first we read the shapes metadata
        #
        query = cls.__context.sparql_context
        query += f"""
        SELECT ?version
        FROM {cls.__graph}:shacl
        WHERE {{
            {cls.__graph}:shapes dcterms:hasVersion ?version .
        }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context=cls.__context, query_result=jsonobj)
        cls.__version = SemanticVersion.fromString(res[0]['version'])
        #
        # now we read the OWL ontology metadata
        #
        query = cls.__context.sparql_context
        query += f"""
        SELECT ?version
        FROM {cls.__graph}:onto
        WHERE {{
            {cls.__graph}:ontology owl:versionInfo ?version .
        }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context=cls.__context, query_result=jsonobj)
        version = SemanticVersion.fromString(res[0]['version'])
        if version != cls.__version:
            raise OldapErrorInconsistency(f'Versionnumber of SHACL ({cls.__version}) and OWL ({version}) do not match')
        #
        # now get the QNames of all standalone properties within the data model
        #
        query = cls.__context.sparql_context
        query += f"""
        SELECT ?prop
        FROM {cls.__graph}:shacl
        WHERE {{
            ?prop a sh:PropertyShape
        }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context=cls.__context, query_result=jsonobj)
        propclasses = []
        for r in res:
            propnameshacl = str(r['prop'])
            propclassiri = propnameshacl.removesuffix("Shape")
            propclass = PropertyClass.read(con, project, Iri(propclassiri, validate=False))
            propclasses.append(propclass)
        #
        # now get all resources defined in the data model
        #
        query = cls.__context.sparql_context
        query += f"""
        SELECT ?shape
        FROM {cls.__graph}:shacl
        WHERE {{
            ?shape a sh:NodeShape
        }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context=cls.__context, query_result=jsonobj)
        resclasses = []
        for r in res:
            resnameshacl = str(r['shape'])
            resclassiri = resnameshacl.removesuffix("Shape")
            resclass = ResourceClass.read(con, project, Iri(resclassiri, validate=False))
            resclasses.append(resclass)
        instance = cls(project=project, con=con, propclasses=propclasses, resclasses=resclasses)
        for qname in instance.get_propclasses():
            instance[qname].set_notifier(instance.notifier, qname)
        for qname in instance.get_resclasses():
            instance[qname].set_notifier(instance.notifier, qname)
        return instance

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        timestamp = Xsd_dateTime.now()
        blank = ''
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context

        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'

        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {self.__graph}:shacl {{\n'

        sparql += f'{blank:{(indent + 2) * indent_inc}}{self.__graph}:shapes dcterms:hasVersion {self.__version.toRdf} .\n'
        sparql += '\n'

        for propiri, propclass in self.__propclasses.items():
            if propclass.internal:
                raise OldapErrorInconsistency(f"Property class {propclass.property_class_iri} is internal and cannot be used here.")
            sparql += propclass.create_shacl(timestamp=timestamp, indent=2)
            sparql += '\n'

        for resiri, resclass in self.__resclasses.items():
            sparql += resclass.create_shacl(timestamp=timestamp, indent=2)
            sparql += '\n'

        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += '\n'

        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {self.__graph}:onto {{\n'

        sparql += f'{blank:{(indent + 2) * indent_inc}}{self.__graph}:ontology owl:type owl:Ontology ;\n'
        sparql += f'{blank:{(indent + 2) * indent_inc}}owl:versionInfo {self.__version.toRdf} .\n'
        #sparql += f'{blank:{(indent + 2) * indent_inc}}owl:imports <http://www.w3.org/2004/02/skos/core> ;\n'
        #sparql += f'{blank:{(indent + 2) * indent_inc}}owl:imports oldap:ontology .\n'
        sparql += '\n'

        for propiri, propclass in self.__propclasses.items():
            sparql += propclass.create_owl_part1(timestamp=timestamp, indent=2)

        for resiri, resclass in self.__resclasses.items():
            sparql += resclass.create_owl(timestamp=timestamp, indent=2)

        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'

        try:
            self._con.transaction_start()
            self._con.transaction_update(sparql)
            for propiri, propclass in self.__propclasses.items():
                propclass.set_creation_metadata(timestamp=timestamp)
            for resiri, resclass in self.__resclasses.items():
                resclass.set_creation_metadata(timestamp=timestamp)
            self._con.transaction_commit()
        except OldapError as err:
            self._con.transaction_abort()
            raise

    def update(self):
        for qname, change in self.__propclasses_changeset.items():
            match(change.action):
                case Action.CREATE:
                    self.__propclasses[qname].create()
                case Action.MODIFY:
                    self.__propclasses[qname].update()
                case Action.DELETE:
                    self.__propclasses[qname].delete()
        for qname, change in self.__resclasses_changeset.items():
            match (change.action):
                case Action.CREATE:
                    self.__resclasses[qname].create()
                case Action.MODIFY:
                    self.__resclasses[qname].update()
                case Action.DELETE:
                    self.__resclasses[qname].delete()

    def write_as_trig(self, filename: str, indent: int = 0, indent_inc: int = 4):
        with open(filename, 'w') as f:
            timestamp = Xsd_dateTime.now()
            blank = ''
            context = Context(name=self._con.context_name)
            f.write('\n')
            f.write(context.turtle_context)
            f.write(f'\n{blank:{indent * indent_inc}}{self.__graph}:shacl {{\n')
            f.write(f'{blank:{(indent + 2) * indent_inc}}{self.__graph}:shapes dcterms:hasVersion {self.__version.toRdf} .\n')
            f.write('\n')
            for iri, prop in self.__propclasses.items():
                if not prop.internal:
                    f.write(prop.create_shacl(timestamp=timestamp, indent=1))
            for iri, resclass in self.__resclasses.items():
                f.write(resclass.create_shacl(timestamp=timestamp, indent=1))
            f.write(f'\n{blank:{indent * indent_inc}}}}\n')

            f.write(f'{blank:{indent * indent_inc}}{self.__graph}:onto {{\n')
            f.write(f'{blank:{(indent + 2) * indent_inc}}{self.__graph}:ontology owl:type owl:Ontology ;\n')
            f.write(f'{blank:{(indent + 2) * indent_inc}}owl:versionInfo {self.__version.toRdf} .\n')
            # f.write(f'{blank:{(indent + 2) * indent_inc}}owl:imports <http://www.w3.org/2004/02/skos/core> ;\n')
            # f.write(f'{blank:{(indent + 2) * indent_inc}}owl:imports oldap:ontology .\n')
            f.write('\n')
            for iri, prop in self.__propclasses.items():
                f.write(prop.create_owl_part1(timestamp=timestamp, indent=2))
            for iri, resclass in self.__resclasses.items():
                f.write(resclass.create_owl(timestamp=timestamp))
            f.write(f'{blank:{indent * indent_inc}}}}\n')







import re
from typing import Tuple, Self

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_gMonthDay(Xsd):
    """
    Implementation of the XML Schema [xsd:gMonthDay](https://www.w3.org/TR/xmlschema11-2/#gMonthDay) datatype
    """
    __month: int
    __day: int
    __tz: Tuple[int, int] | None
    __zulu: bool

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructor of the Xsd_gMonthDay class
        :param value: Xsd_gMonthDay instance or valid string
        :type value: Xsd_gMonthDay | str
        :raises OldapErrorValue: if the value is not valid Xsd_gMonthDay representation
        """
        if isinstance(value, Xsd_gMonthDay):
            self.__month = value.__month
            self.__day = value.__day
            self.__tz = value.__tz
        else:
            if not XsdValidator.validate(XsdDatatypes.gMonthDay, value):
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonthDay')
            res = re.split("--([0-9]{2})-([0-9]{2})((([+-][0-9]{2}):([0-9]{2}))|(Z))?", value)
            if len(res) != 9:
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonthDay.')
            self.__month = int(res[1])
            self.__day = int(res[2])
            if res[3] == 'Z':
                self.__tz = (0, 0)
                self.__zulu = True
            elif res[3] is not None:
                self.__tz = (int(res[5]), int(res[6]))
                self.__zulu = False
            else:
                self.__tz = None
        if self.__month < 1 or self.__month > 12 or self.__day < 1 or self.__day > 31:
            raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonthDay.')

    def __str__(self):
        """
        String representation of the Xsd_gMonthDay instance
        :return: String representation of the Xsd_gMonthDay instance
        """
        s = f'--{self.__month:02}-{self.__day:02}'
        if self.__tz is not None:
            if self.__zulu:
                s += 'Z'
            else:
                s += f'{self.__tz[0]:0=+3}:{self.__tz[1]:02}'
        return s

    def __repr__(self):
        """
        Constructor string representation of the Xsd_gMonthDay instance
        :return: Constructor string
        """
        return f'Xsd_gMonthDay("{str(self)}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Equality check for Xsd_gMonthDay instance
        :param other: Another Xsd_gMonthDay instance or a valid string
        :type other: Xsd_gMonthDay | str | None
        :return: True or False
        :raise OldapErrorValue: if the value is not valid Xsd_gMonthDay representation
        """
        if other is None:
            return False
        if not isinstance(other, Xsd_gMonthDay):
            other = Xsd_gMonthDay(other)
        if self.__month != other.__month:
            return False
        if self.__day != other.__day:
            return False
        if self.__tz != other.__tz:
            return False
        return True

    def __hash__(self) -> int:
        """
        Hash value of the Xsd_gMonthDay instance
        :return: Hash value
        """
        return hash(str(self))

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method for JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': str(self)}

    @property
    def toRdf(self) -> str:
        """
        RDF representation of the Xsd_gMonthDay instance
        :return: RDF string
        """
        return f'"{str(self)}"^^xsd:gMonthDay'



from typing import Set, List, Dict, Iterable, Iterator, Self, TypeVar, Generic

from pystrict import strict

from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorType
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd

T = TypeVar("T")


@serializer
class RdfSet(Generic[T], Notify):
    __data: Set[T]

    def __init__(self, *args: Self | set[T] | list[T] | tuple[T] | T,
                 value: Self | set[T] | list[T] | tuple[T] | T | None = None) -> None:
        self.__data: Set[T] = set()
        Notify.__init__(self)
        if len(args) == 0:
            if value is None:
                return
            else:
                if isinstance(value, RdfSet):
                    self.__data = value.__data
                elif isinstance(value, (set | list | tuple)):
                    for val in value:
                        self.__data.add(val)
                else:
                    self.__data.add(value)
        elif len(args) == 1:
            if isinstance(args[0], RdfSet):
                self.__data = args[0].__data
            elif isinstance(args[0], (set | list | tuple)):
                for val in args[0]:
                    self.__data.add(val)
            else:
                self.__data.add(args[0])
        else:
            for val in args:
                self.__data.add(val)

    def __len__(self) -> int:
        return len(self.__data)

    def __eq__(self, other: Self | set[T] | None) -> bool:
        if other is None:
            return False
        if isinstance(other, RdfSet):
            return self.__data == other.__data
        elif isinstance(other, set):
            return self.__data == other
        else:
            raise OldapErrorValue(f"Comparison between RdfSet and {type(other)} not possible")

    def __gt__(self, other: Self | set[T]) -> bool:
        if isinstance(other, RdfSet):
            return self.__data > other.__data
        if isinstance(other, set):
            return self.__data > other
        raise OldapErrorType(f'Cannot compare {type(self).__name__} to {type(other).__name__}')

    def __ge__(self, other: Self | set[T]) -> bool:
        if isinstance(other, RdfSet):
            return self.__data >= other.__data
        if isinstance(other, set):
            return self.__data >= other
        raise OldapErrorType(f'Cannot compare {type(self).__name__} to {type(other).__name__}')

    def __lt__(self, other: Self | set[T]) -> bool:
        if isinstance(other, RdfSet):
            return self.__data < other.__data
        if isinstance(other, set):
            return self.__data < other
        raise OldapErrorType(f'Cannot compare {type(self).__name__} to {type(other).__name__}')

    def __le__(self, other: Self | set[T]) -> bool:
        if isinstance(other, RdfSet):
            return self.__data <= other.__data
        if isinstance(other, set):
            return self.__data <= other
        raise OldapErrorType(f'Cannot compare {type(self).__name__} to {type(other).__name__}')

    def __str__(self) -> str:
        return '(' + ", ".join(map(str, self.__data)) + ')'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(' + ", ".join(map(repr, self.__data)) + ')'

    def __contains__(self, val: T) -> bool:
        return val in self.__data

    def __iter__(self) -> Iterator[T]:
        return iter(self.__data)

    def add(self, val: T) -> None:
        self.notify()
        self.__data.add(val)

    def discard(self, val: T) -> None:
        self.notify()
        self.__data.discard(val)

    @property
    def value(self) -> set[T]:
        return self.__data

    def _as_dict(self) -> Dict[str, List[T]]:
        return {'value': [x for x in self.__data]}

    @property
    def toRdf(self) -> str:
        return f'({" ".join(map(lambda x: x.toRdf, self.__data))})'

    @property
    def value(self) -> set[T]:
        return self.__data

if __name__ == "__main__":
    g = RdfSet[str](['a', 'b', 'c'])
    print(repr(g))



from dataclasses import dataclass
from enum import Enum, unique
from functools import partial
from typing import Self

from oldaplib.src.enums.action import Action
from oldaplib.src.enums.permissions import AdminPermission
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorInconsistency, OldapErrorImmutable, OldapError, OldapErrorNoPermission, OldapErrorAlreadyExists, \
    OldapErrorNotFound, OldapErrorUpdateFailed, OldapErrorInUse
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string

OldapListAttrTypes = LangString | Iri | None

@dataclass
class OldapListAttrChange:
    """
    A dataclass used to represent the changes made to a field.
    """
    old_value: OldapListAttrTypes
    action: Action


@unique
class OldapListAttr(Enum):
    """
    This enum class represents the fields used in the project model
    """
    OLDAPLIST_IRI = 'oldap:oldapListIri'  # virtual property, repents the RDF subject
    PREF_LABEL = 'skos:prefLabel'
    DEFINITION = 'skos:definition'

class OldapList(Model):

    __datatypes = {
        OldapListAttr.OLDAPLIST_IRI: Iri,
        OldapListAttr.PREF_LABEL: LangString,
        OldapListAttr.DEFINITION: LangString,
    }

    __creator: Iri | None
    __created: Xsd_dateTime | None
    __contributor: Iri | None
    __modified: Xsd_dateTime | None
    __project: Project
    __graph: Xsd_NCName

    __attributes: dict[OldapListAttr, OldapListAttrTypes]

    __changeset: dict[OldapListAttr, OldapListAttrChange]

    def __init__(self, *,
                 con: IConnection,
                 project: Project,
                 creator: Iri | None = None,
                 created: Xsd_dateTime | None = None,
                 contributor: Iri | None = None,
                 modified: Xsd_dateTime | None = None,
                 oldapListIri: Iri | str | None = None,
                 prefLabel: LangString | str | None = None,
                 definition: LangString | str | None = None):
        super().__init__(con)
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        self.__project = project
        context = Context(name=self._con.context_name)
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        self.__graph = project.projectShortName

        self.__creator = creator if creator is not None else con.userIri
        if created and not isinstance(created, Xsd_dateTime):
            raise OldapErrorValue(f'Created must be "Xsd_dateTime", not "{type(created)}".')
        self.__created = created
        self.__contributor = contributor if contributor is not None else con.userIri
        if modified and not isinstance(modified, Xsd_dateTime):
            raise OldapErrorValue(f'Modified must be "Xsd_dateTime", not "{type(modified)}".')
        self.__modified = modified
        self.__attributes = {}

        if oldapListIri:
            if not isinstance(oldapListIri, Iri):
                self.__attributes[OldapListAttr.OLDAPLIST_IRI] = Iri(oldapListIri)
            else:
                self.__attributes[OldapListAttr.OLDAPLIST_IRI] = oldapListIri
        else:
            self.__attributes[OldapListAttr.OLDAPLIST_IRI] = Iri()

        self.__attributes[OldapListAttr.PREF_LABEL] = prefLabel if isinstance(prefLabel, LangString) else LangString(prefLabel)
        self.__attributes[OldapListAttr.PREF_LABEL].set_notifier(self.notifier, Iri(OldapListAttr.PREF_LABEL.value))
        self.__attributes[OldapListAttr.DEFINITION] = definition if isinstance(definition, LangString) else LangString(definition)
        self.__attributes[OldapListAttr.DEFINITION].set_notifier(self.notifier, Iri(OldapListAttr.DEFINITION.value))

        #
        # Consistency checks
        #
        if not self.__attributes[OldapListAttr.PREF_LABEL]:
            raise OldapErrorInconsistency(f'Project must have at least one skos:prefLabel, none given.')

        #
        # create all the attributes of the class according to the OldapListAttr definition
        #
        for attr in OldapListAttr:
            prefix, name = attr.value.split(':')
            setattr(OldapList, name, property(
                partial(OldapList.__get_value, attr=attr),
                partial(OldapList.__set_value, attr=attr),
                partial(OldapList.__del_value, attr=attr)))
        self.__changeset = {}

    def check_for_permissions(self) -> (bool, str):
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            #
            # user has root privileges!
            #
            return True, "OK – IS ROOT"
        else:
            if len(self.inProject) == 0:
                return False, f'Actor has no ADMIN_LISTS permission for user {self.userId}.'
            allowed: list[Iri] = []
            for proj in self.inProject.keys():
                if actor.inProject.get(proj) is None:
                    return False, f'Actor has no ADMIN_LISTS permission for project {proj}'
                else:
                    if AdminPermission.ADMIN_LISTS not in actor.inProject.get(proj):
                        return False, f'Actor has no ADMIN_LISTS permission for project {proj}'
            return True, "OK"

    def __get_value(self: Self, attr: OldapListAttr) -> OldapListAttrTypes | None:
        return self.__attributes.get(attr)

    def __set_value(self: Self, value: OldapListAttrTypes, attr: OldapListAttr) -> None:
        self.__change_setter(attr, value)

    def __del_value(self: Self, attr: OldapListAttr) -> None:
        self.__changeset[attr] = OldapListAttrChange(self.__attributes[attr], Action.DELETE)
        del self.__attributes[attr]

    def __change_setter(self, attr: OldapListAttr, value: OldapListAttrTypes) -> None:
        if self.__attributes.get(attr) == value:
            return
        if attr == OldapListAttr.OLDAPLIST_IRI:
            raise OldapErrorImmutable(f'Field {attr.value} is immutable.')
        if self.__attributes.get(attr) is None:
            if self.__changeset.get(attr) is None:
                self.__changeset[attr] = OldapListAttrChange(None, Action.CREATE)
        else:
            if value is None:
                if self.__changeset.get(attr) is None:
                    self.__changeset[attr] = OldapListAttrChange(self.__attributes[attr], Action.DELETE)
            else:
                if self.__changeset.get(attr) is None:
                    self.__changeset[attr] = OldapListAttrChange(self.__attributes[attr], Action.REPLACE)
        if value is None:
            del self.__attributes[attr]
        else:
            if not isinstance(value, self.__datatypes[attr]):
                self.__attributes[attr] = self.__datatypes[attr](value)
            else:
                self.__attributes[attr] = value

    def __str__(self):
        res = f'OldapList: {self.__attributes[OldapListAttr.OLDAPLIST_IRI]}\n'\
              f'  Creation: {self.__created} by {self.__creator}\n'\
              f'  Modified: {self.__modified} by {self.__contributor}\n'\
              f'  Preferred label: {self.__attributes.get(OldapListAttr.PREF_LABEL)}\n'\
              f'  Definition: {self.__attributes.get(OldapListAttr.DEFINITION)}'
        return res

    def __getitem__(self, attr: OldapListAttr) -> OldapListAttrTypes:
        return self.__attributes[attr]

    def get(self, attr: OldapListAttr) -> OldapListAttrTypes:
        return self.__attributes.get(attr)

    def __setitem__(self, attr: OldapListAttr, value: OldapListAttrTypes) -> None:
        self.__change_setter(attr, value)

    def __delitem__(self, attr: OldapListAttr) -> None:
        if self.__attributes.get(attr) is not None:
            self.__changeset[attr] = OldapListAttrChange(self.__attributes[attr], Action.DELETE)
            del self.__attributes[attr]


    @property
    def creator(self) -> Iri | None:
        """
        The creator of the OldapList.
        :return: Iri of the creator of the OldapList.
        :rtype: Iri | None
        """
        return self.__creator

    @property
    def created(self) -> Xsd_dateTime | None:
        """
        The creation date of the OldapList.
        :return: Creation date of the OldapList.
        :rtype: Xsd_dateTime | None
        """
        return self.__created

    @property
    def contributor(self) -> Iri | None:
        """
        The contributor of the OldapList as Iri.
        :return: Iri of the contributor of the OldapList.
        :rtype: Iri | None
        """
        return self.__contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        """
        Modification date of the OldapList.
        :return: Modification date of the OldapList.
        :rtype: Xsd_dateTime | None
        """
        return self.__modified

    @property
    def changeset(self) -> dict[OldapListAttr, OldapListAttrChange]:
        """
        Return the changeset, that is dicst with information about all properties that have benn changed.
        This method is only for internal use or debugging...
        :return: A dictionary of all changes
        :rtype: Dict[ProjectAttr, ProjectAttrChange]
        """
        return self.__changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset. This method is only for internal use or debugging...
        :return: None
        """
        self.__changeset = {}

    def notifier(self, attrname: Iri) -> None:
        """
        This method is called when a field is being changed.
        :param fieldname: Fieldname of the field being modified
        :return: None
        """
        attr = OldapListAttr(attrname)
        self.__changeset[attr] = OldapListAttrChange(self.__attributes[attr], Action.MODIFY)

    @classmethod
    def read(cls, con: IConnection, project: Project, oldapListIri: Iri | str) -> Self:
        oldapListIri = Iri(oldapListIri)

        context = Context(name=con.context_name)
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        graph = project.projectShortName

        query = context.sparql_context
        query += f"""
            SELECT ?prop ?val
            FROM {graph}:lists
            WHERE {{
                {oldapListIri.toRdf} ?prop ?val
            }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'OldapList with IRI "{oldapListIri}" not found.')
        creator: Iri | None = None
        created: Xsd_dateTime | None = None
        contributor: Iri | None = None
        modified: Xsd_dateTime | None = None
        oldapList: Iri | None = None
        prefLabel: LangString | None = None
        definition: LangString | None = None
        for r in res:
            match str(r.get('prop')):
                case 'dcterms:creator':
                    creator = r['val']
                case 'dcterms:created':
                    created = r['val']
                case 'dcterms:contributor':
                    contributor = r['val']
                case 'dcterms:modified':
                    modified = r['val']
                case OldapListAttr.PREF_LABEL.value:
                    if not prefLabel:
                        prefLabel = LangString()
                    prefLabel.add(r['val'])
                case OldapListAttr.DEFINITION.value:
                    if not definition:
                        definition = LangString()
                    definition.add(r['val'])
        return cls(con=con,
                   project=project,
                   creator=creator,
                   created=created,
                   contributor=contributor,
                   modified=modified,
                   oldapListIri=oldapListIri,
                   prefLabel=prefLabel,
                   definition=definition)

    @staticmethod
    def search(con: IConnection,
               project: Project,
               prefLabel: Xsd_string | str | None = None,
               definition: str | None = None) -> list[Iri]:
        context = Context(name=con.context_name)
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        graph = project.projectShortName

        prefLabel = Xsd_string(prefLabel)
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        graph = project.projectShortName
        sparql = context.sparql_context
        sparql += 'SELECT DISTINCT ?list\n'
        sparql += f'FROM {graph}:lists\n'
        sparql += 'WHERE {\n'
        sparql += '   ?list a oldap:OldapList .\n'
        if prefLabel:
            sparql += '   ?list skos:prefLabel ?label .\n'
            if prefLabel.lang:
                sparql += f'   FILTER(?label = "{prefLabel.toRdf}")\n'
            else:
                sparql += f'   FILTER(STR(?label) = "{Xsd_string.escaping(prefLabel.value)}")\n'
        if definition:
            sparql += '   ?list skos:definition ?definition .\n'
            sparql += f'   FILTER(CONTAINS(STR(?definition), "{Xsd_string.escaping(definition.value)}"))\n'
        sparql += '}\n'

        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        lists: list[Iri] = []
        if len(res) > 0:
            for r in res:
                lists.append(r['list'])
        return lists

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        if self._con is None:
            raise OldapError("Cannot create: no connection")
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        indent: int = 0
        indent_inc: int = 4

        context = Context(name=self._con.context_name)

        sparql1 = context.sparql_context
        sparql1 += f"""
        SELECT ?list
        FROM {self.__graph}:lists
        WHERE {{
            ?list a oldap:OldapList .
            FILTER(?list = {self.oldapListIri.toRdf})
        }}
        """

        blank = ''
        sparql2 = context.sparql_context
        sparql2 += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql2 += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH {self.__graph}:lists {{'
        sparql2 += f'\n{blank:{(indent + 2) * indent_inc}}{self.oldapListIri.toRdf} a oldap:OldapList'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}{OldapListAttr.PREF_LABEL.value} {self.prefLabel.toRdf}'
        if self.definition:
            sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}{OldapListAttr.DEFINITION.value} {self.definition.toRdf}'
        sparql2 += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql2 += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            jsonobj = self._con.transaction_query(sparql1)
        except OldapError:
            self._con.transaction_abort()
            raise
        res = QueryProcessor(context, jsonobj)
        if len(res) > 0:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A list with a oldapListIri "{self.oldapListIri}" already exists')

        try:
            self._con.transaction_update(sparql2)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__created = timestamp
        self.__creator = self._con.userIri
        self.__modified = timestamp
        self.__contributor = self._con.userIri

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        blank = ''
        sparql_list = []

        for field, change in self.__changeset.items():
            if field == OldapListAttr.PREF_LABEL or field == OldapListAttr.DEFINITION:
                if change.action == Action.MODIFY:
                    sparql_list.extend(self.__attributes[field].update(graph=Xsd_QName(f'{self.__graph}:lists'),
                                                                       subject=self.oldapListIri,
                                                                       subjectvar='?list',
                                                                       field=Xsd_QName(field.value)))
                if change.action == Action.DELETE or change.action == Action.REPLACE:
                    # sparql = self.__attributes[field].delete(graph=Xsd_QName(f'{self.__graph}:lists'),
                    #                                          subject=self.oldapListIri,
                    #                                          field=Xsd_QName(field.value))
                    sparql = self.__changeset[field].old_value.delete(graph=Xsd_QName(f'{self.__graph}:lists'),
                                                                      subject=self.oldapListIri,
                                                                      field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                if change.action == Action.CREATE or change.action == Action.REPLACE:
                    sparql = self.__attributes[field].create(graph=Xsd_QName(f'{self.__graph}:lists'),
                                                             subject=self.oldapListIri,
                                                             field=Xsd_QName(field.value))
                    sparql_list.append(sparql)

        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)

        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName(f'{self.__graph}:lists'), self.oldapListIri, self.modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName(f'{self.__graph}:lists'), self.oldapListIri)
        except OldapError:
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed("Update failed! Timestamp does not match")
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__modified = timestamp
        self.__contributor = self._con.userIri  # TODO: move creator, created etc. to Model!

    def delete(self) -> None:
        """
        Delete the given user from the triplestore
        :return: None
        :raises OldapErrorNoPermission: No permission for operation
        :raises OldapError: generic internal error
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f'''
        SELECT ?listnode
        FROM {self.__graph}:lists
        WHERE {{
            ?listnode a oldap:OldapListNode .
        }}
        '''
        jsonobj = self._con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) > 0:
            raise OldapErrorInUse(f'List {self.prefLabel} cannot be deleted since there are still nodes.')

        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            {self.oldapListIri.toRdf} a oldap:OldapList .
            {self.oldapListIri.toRdf} ?prop ?val .
        }} 
        """
        # TODO: use transaction for error handling
        self._con.update_query(sparql)






import uuid
from dataclasses import dataclass
from enum import unique, Enum
from functools import partial

from pystrict import strict
from typing import List, Dict, Optional, Self
from datetime import date, datetime

from oldaplib.src.enums.permissions import AdminPermission
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.helpers.tools import lprint
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorValue, OldapErrorAlreadyExists, OldapErrorNoPermission, \
    OldapErrorUpdateFailed, OldapErrorImmutable, OldapErrorNotFound, OldapErrorInconsistency
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.xsd.xsd_string import Xsd_string

ProjectAttrTypes = LangString | Xsd | None

@dataclass
class ProjectAttrChange:
    """
    A dataclass used to represent the changes made to a field.
    """
    old_value: ProjectAttrTypes
    action: Action

@unique
class ProjectAttr(Enum):
    """
    This enum class represents the fields used in the project model
    """
    PROJECT_IRI = 'oldap:projectIri'  # virtual property, repents the RDF subject
    PROJECT_SHORTNAME = 'oldap:projectShortName'
    LABEL = 'rdfs:label'
    COMMENT = 'rdfs:comment'
    NAMESPACE_IRI = 'oldap:namespaceIri'
    PROJECT_START = 'oldap:projectStart'
    PROJECT_END = 'oldap:projectEnd'

#@strict
class Project(Model):
    """
    # Project

    This class implements the Project model. A Project is a distinct research space within Oldap
    framework that offers dedicated space for data, its own data modeling and access control. A project
    needs the following metadata:

    - `projectIri: Iri | None` [optional]:
      The IRI that uniquely identifies this project. This can be an
      [Iri](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.Iri) instance or a string with either an Iri or Qname
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.QName). If the `projectIri` is ommitted,
      a random IRI based on the URN semantics will be generated.
    - `projectShortName: Xsd_NCName | str` [mandatory]:
      The short name of the project that must be a NCName or a string
    - `label: LangString | None` [optional]:
      A multilingual string with a human-readable label for the project (`rdfs:label`)
    - `comment: LangString | None` [optional]:
      A multilingual description of the project (`rdfs:comment`)
    - `namespaceIri: NamespaceIri` [mandatory]:
       The namespace that the project uses for its data and data model. Must be
       a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart: Xsd_date | date | None` [optional]:
      The start date of the project.  Must be a Python `date` type. If not set,
       the current date will be used.
    - `projectEnd: Xsd_date | None` [optional]:
      The optional end date of the project. Must be a Xsd_date or a Python `date` type.

    The class provides the following methods:

    - [Project(...)](/python_docstrings/project#oldaplib.src.project.Project.__init__)): Constructor
    - [read(...)](/python_docstrings/project#oldaplib.src.project.Project.read):
      Read project form triplestore and return a Project-instance
    - [search(...)](/python_docstrings/project#oldaplib.src.project.Project.search):
      Search a Project in the triplestore
    - [create(...)](/python_docstrings/project#oldaplib.src.project.Project.create): Create a new project in the triplestore
    - [update(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Write changes to triplestore
    - [delete(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Delete a project in the triplestore

    The class provides the following properties:

    - `projectIri` [read only]: The project IRI
    - `projectShortName` [read/write]: The project short name. Must be a NCName
    - `projectLabel` [read/write]: The projects label as multilingual LangString
    - `projectComment` [read/write]: The projects comment/description as multilingual
      [LangString](/python_docstrings/langstring)
    - `namespaceIri` [read only]: The namespace that the project uses for its data and data model. Must be
      a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart` [read/write]: The start date of the project. Must be a Python `date` type
    - `projectEnd` [read/write]: The end date of the project. Must be a Python `date` type
    - `creator` [read only]: The creator of the project. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `created` [read only]: The creation date of the project. Must be a Python `date` type
    - `contributor` [read only]: The person which made the last changes to the project data. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `modified` [read only]: The modification date of the project. Must be a Python `date` type

    """
    __datatypes = {
        ProjectAttr.PROJECT_IRI: Iri,
        ProjectAttr.PROJECT_SHORTNAME: Xsd_NCName,
        ProjectAttr.LABEL: LangString,
        ProjectAttr.COMMENT: LangString,
        ProjectAttr.NAMESPACE_IRI: NamespaceIRI,
        ProjectAttr.PROJECT_START: Xsd_date,
        ProjectAttr.PROJECT_END: Xsd_date,
    }

    __creator: Iri | None
    __created: Xsd_dateTime | None
    __contributor: Iri | None
    __modified: Xsd_dateTime | None

    __attributes: dict[ProjectAttr, ProjectAttrTypes]

    __changeset: dict[ProjectAttr, ProjectAttrChange]

    def __init__(self, *,
                 con: IConnection,
                 creator: Iri | None = None,
                 created: Xsd_dateTime | None = None,
                 contributor: Iri | None = None,
                 modified: Xsd_dateTime | None = None,
                 projectIri: Iri | str | None = None,
                 projectShortName: Xsd_NCName | str,
                 namespaceIri: NamespaceIRI,
                 label: LangString | str | None = None,
                 comment: LangString | str | None = None,
                 projectStart: Xsd_date | None = None,
                 projectEnd: Xsd_date | None = None):
        """
        Constructs a new Project
        :param con: [Connection](/python_docstrings/iconnection) instance
        :param creator: Creator of the project  [DO NOT SET!]
        :type creator: Xsd_anyURI | None
        :param created: Date the project was created  [DO NOT SET!]
        :type created: datetime | None
        :param contributor: person that made the last change  [DO NOT SET!]
        :type contributor: Xsd_anyURI
        :param modified: Last date the project was modified  [DO NOT SET!]
        :type modified: date | None
        :param projectIri: IRI to be used for the project. If no projectIRI is provied, the constrctor
         will create an arbitrary IRI based on thr URN scheme and a UUID. [optional].
        :type projectIri: Iri
        :param projectShortName: A short name for the project. Is used as prefix for named graphs that
           are being used for the project. [mandatory]
        :type projectShortName: NCname (strings are accepted only if conform to NCName syntax)
        :param namespaceIri: The namespace to be used for the projects data [mandatory]
        :type namespaceIri: NamespaceIRI
        :param label: Human-readable name for the project (multi-language) (`rdfs:label`) [optional]
        :type label: LangString
        :param comment: Description of the project (multi-language) (`rdfs:comment`)
        :type comment: LangString
        :param projectStart: Start date of the project
        :type projectStart: Python date
        :param projectEnd: End date of the project [Optional]
        :type projectEnd: Python date
        :raises OldapErrorValue: Invalid parameter supplied
        """
        super().__init__(con)
        self.__creator = creator if creator is not None else con.userIri
        if created and not isinstance(created, Xsd_dateTime):
            raise OldapErrorValue(f'Created must be "Xsd_dateTime", not "{type(created)}".')
        self.__created = created
        self.__contributor = contributor if contributor is not None else con.userIri
        if modified and not isinstance(modified, Xsd_dateTime):
            raise OldapErrorValue(f'Modified must be "Xsd_dateTime", not "{type(modified)}".')
        self.__modified = modified
        self.__attributes = {}

        if projectIri:
            if not isinstance(projectIri, Iri):
                self.__attributes[ProjectAttr.PROJECT_IRI] = Iri(projectIri)
            else:
                self.__attributes[ProjectAttr.PROJECT_IRI] = projectIri
        else:
            self.__attributes[ProjectAttr.PROJECT_IRI] = Iri()

        self.__attributes[ProjectAttr.PROJECT_SHORTNAME] = projectShortName if isinstance(projectShortName, Xsd_NCName) else Xsd_NCName(projectShortName)

        if namespaceIri and isinstance(namespaceIri, NamespaceIRI):
            self.__attributes[ProjectAttr.NAMESPACE_IRI] = namespaceIri
        else:
            raise OldapErrorValue(f'Invalid namespace iri: {namespaceIri}')

        self.__attributes[ProjectAttr.LABEL] = label if isinstance(label, LangString) else LangString(label)
        self.__attributes[ProjectAttr.LABEL].set_notifier(self.notifier, Xsd_QName(ProjectAttr.LABEL.value))
        self.__attributes[ProjectAttr.COMMENT] = comment if isinstance(comment, LangString) else LangString(comment)
        self.__attributes[ProjectAttr.COMMENT].set_notifier(self.notifier, Xsd_QName(ProjectAttr.COMMENT.value))
        self.__attributes[ProjectAttr.PROJECT_SHORTNAME] = projectShortName if isinstance(projectShortName, Xsd_NCName) else Xsd_NCName(projectShortName)
        #
        # Consistency checks
        #
        if not self.__attributes[ProjectAttr.LABEL]:
            raise OldapErrorInconsistency(f'Project must have at least one rdfs:label, none given.')
        if projectStart is not None:
            self.__attributes[ProjectAttr.PROJECT_START] = projectStart if isinstance(projectStart, Xsd_date) else Xsd_date(projectStart)
        else:
            self.__attributes[ProjectAttr.PROJECT_START] = Xsd_date()
        if projectEnd is not None:
            self.__attributes[ProjectAttr.PROJECT_END] = projectEnd if isinstance(projectEnd, Xsd_date) else Xsd_date(projectEnd)
            if self.__attributes[ProjectAttr.PROJECT_END] < self.__attributes[ProjectAttr.PROJECT_START]:
                raise OldapErrorInconsistency(f'Project start date {projectStart} is after project end date {projectEnd}.')

        #
        # create all the attributes of the class according to the ProjectFields dfinition
        #
        for attr in ProjectAttr:
            prefix, name = attr.value.split(':')
            setattr(Project, name, property(
                partial(Project.__get_value, attr=attr),
                partial(Project.__set_value, attr=attr),
                partial(Project.__del_value, attr=attr)))
        self.__changeset = {}

    def check_for_permissions(self) -> (bool, str):
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            return True, "OK"
        else:
            return False, "No permission to create a new project."

    def __get_value(self: Self, attr: ProjectAttr) -> ProjectAttrTypes | None:
        tmp = self.__attributes.get(attr)
        if not tmp:
            return None
        return tmp

    def __set_value(self: Self, value: ProjectAttrTypes, attr: ProjectAttr) -> None:
        self.__change_setter(attr, value)

    def __del_value(self: Self, attr: ProjectAttr) -> None:
        self.__changeset[attr] = ProjectAttrChange(self.__attributes[attr], Action.DELETE)
        del self.__attributes[attr]

    #
    # this private method handles the setting of a field. Whenever a field is being
    # set or modified, this method is called. It also puts the original value and the
    # action into the changeset.
    #
    def __change_setter(self, field: ProjectAttr, value: ProjectAttrTypes) -> None:
        if self.__attributes.get(field) == value:
            return
        if field == ProjectAttr.PROJECT_IRI or field == ProjectAttr.NAMESPACE_IRI or field == ProjectAttr.PROJECT_SHORTNAME:
            raise OldapErrorImmutable(f'Field {field.value} is immutable.')
        if field == ProjectAttr.PROJECT_START:
            if self.__attributes.get(ProjectAttr.PROJECT_END) and value >= self.__attributes[ProjectAttr.PROJECT_END]:
                raise OldapErrorInconsistency('Project start date must be less than project end date.')
        if field == ProjectAttr.PROJECT_END:
            if self.__attributes.get(ProjectAttr.PROJECT_START) and value <= self.__attributes[ProjectAttr.PROJECT_START]:
                raise OldapErrorInconsistency('Project end date must be greater than project start date.')
        if self.__attributes.get(field) is None:
            if self.__changeset.get(field) is None:
                self.__changeset[field] = ProjectAttrChange(None, Action.CREATE)
        else:
            if value is None:
                if self.__changeset.get(field) is None:
                    self.__changeset[field] = ProjectAttrChange(self.__attributes[field], Action.DELETE)
            else:
                if self.__changeset.get(field) is None:
                    self.__changeset[field] = ProjectAttrChange(self.__attributes[field], Action.REPLACE)
        if value is None:
            del self.__attributes[field]
        else:
            if not isinstance(value, self.__datatypes[field]):
                self.__attributes[field] = self.__datatypes[field](value)
            else:
                self.__attributes[field] = value

    def __str__(self) -> str:
        """
        String representation of the project. This is a multiline string for the human reader.
        :return: str
        """
        res = f'Project: {self.__attributes[ProjectAttr.PROJECT_IRI]}\n'\
              f'  Creation: {self.__created} by {self.__creator}\n'\
              f'  Modified: {self.__modified} by {self.__contributor}\n'\
              f'  Label: {self.__attributes[ProjectAttr.LABEL]}\n'\
              f'  Comment: {self.__attributes[ProjectAttr.COMMENT]}\n'\
              f'  ShortName: {self.__attributes[ProjectAttr.PROJECT_SHORTNAME]}\n'\
              f'  Namespace IRI: {self.__attributes[ProjectAttr.NAMESPACE_IRI]}\n'\
              f'  Project start: {self.__attributes[ProjectAttr.PROJECT_START]}\n'
        if self.__attributes.get(ProjectAttr.PROJECT_END) is not None:
            res += f'  Project end: {self.__attributes[ProjectAttr.PROJECT_END]}\n'
        return res

    def __getitem__(self, attr: ProjectAttr) -> ProjectAttrTypes:
        return self.__attributes[attr]

    def get(self, attr: ProjectAttr) -> ProjectAttrTypes:
        return self.__attributes.get(attr)

    def __setitem__(self, attr: ProjectAttr, value: ProjectAttrTypes) -> None:
        self.__change_setter(attr, value)

    def __delitem__(self, attr: ProjectAttr) -> None:
        if self.__attributes.get(attr) is not None:
            self.__changeset[attr] = ProjectAttrChange(self.__attributes[attr], Action.DELETE)
            del self.__attributes[attr]


    @property
    def creator(self) -> Iri | None:
        """
        The creator of the project.
        :return: Iri of the creator of the project.
        :rtype: Iri | None
        """
        return self.__creator

    @property
    def created(self) -> Xsd_dateTime | None:
        """
        The creation date of the project.
        :return: Creation date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self.__created

    @property
    def contributor(self) -> Iri | None:
        """
        The contributor of the project as Iri.
        :return: Iri of the contributor of the project.
        :rtype: Iri | None
        """
        return self.__contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        """
        Modification date of the project.
        :return: Modification date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self.__modified

    @property
    def changeset(self) -> Dict[ProjectAttr, ProjectAttrChange]:
        """
        Return the changeset, that is dicst with information about all properties that have benn changed.
        This method is only for internal use or debugging...
        :return: A dictionary of all changes
        :rtype: Dict[ProjectAttr, ProjectAttrChange]
        """
        return self.__changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset. This method is only for internal use or debugging...
        :return: None
        """
        self.__changeset = {}

    def notifier(self, fieldname: Xsd_QName) -> None:
        """
        This method is called when a field is being changed.
        :param fieldname: Fieldname of the field being modified
        :return: None
        """
        field = ProjectAttr(fieldname)
        self.__changeset[field] = ProjectAttrChange(self.__attributes[field], Action.MODIFY)

    @classmethod
    def read(cls, con: IConnection, projectIri_SName: Iri | Xsd_NCName | str) -> Self:
        """
        Read the project from the triplestore and return an instance of the project
        :param con: A valid Connection object
        :type con: IConnection
        :param projectIri_SName: The IRI or shortname of the project to be read
        :type projectIri_SName: Iri | Xsd_NCName | str
        :return: Project instance
        :rtype: Project
        :raise: OldapErrorNotFound: project with given Iri not found
        :raise: OldapError: All other errors/problems
        """
        context = Context(name=con.context_name)
        query = context.sparql_context

        projectIri: Iri | None = None
        shortname: Xsd_NCName | None = None
        if isinstance(projectIri_SName, Iri):
            projectIri = projectIri_SName
        elif isinstance(projectIri_SName, Xsd_NCName):
            shortname = Xsd_NCName(projectIri_SName)
        else:
            if ':' in str(projectIri_SName):  # must be IRI or QName
                projectIri = Iri(projectIri_SName)
            else:
                shortname = Xsd_NCName(projectIri_SName)
        if projectIri is not None:
            query += f"""
                SELECT ?prop ?val
                FROM oldap:admin
                WHERE {{
                    {projectIri.toRdf} ?prop ?val
                }}
            """
        elif shortname is not None:
            query += f"""
                SELECT ?proj ?prop ?val
                FROM oldap:admin
                WHERE {{
                    ?proj a oldap:Project .
                    ?proj oldap:projectShortName ?shortname .
                    ?proj ?prop ?val .
                    FILTER(?shortname = {shortname.toRdf})
                }}
            """
        jsonobj = con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'Project with IRI/shortname "{projectIri_SName}" not found.')
        creator: Iri | None = None
        created: Xsd_dateTime | None = None
        contributor: Iri | None = None
        modified: Xsd_dateTime | None = None
        projectShortName = None
        namespaceIri = None
        label = LangString()
        comment = LangString()
        projectStart = None
        projectEnd = None
        for r in res:
            if projectIri is None:
                projectIri = r['proj']
            match str(r.get('prop')):
                case 'dcterms:creator':
                    creator = r['val']
                case 'dcterms:created':
                    created = r['val']
                case 'dcterms:contributor':
                    contributor = r['val']
                case 'dcterms:modified':
                    modified = r['val']
                case 'oldap:namespaceIri':
                    namespaceIri = NamespaceIRI(r['val'])
                case 'oldap:projectShortName':
                    projectShortName = r['val']
                case 'rdfs:label':
                    label.add(str(r['val']))
                case 'rdfs:comment':
                    comment.add(str(r['val']))
                case 'oldap:projectStart':
                    projectStart = r['val']
                case 'oldap:projectEnd':
                    projectEnd = r['val']
        label.changeset_clear()
        label.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.LABEL.value))
        comment.changeset_clear()
        comment.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.COMMENT.value))
        context[projectShortName] = namespaceIri
        return cls(con=con,
                   creator=creator,
                   created=created,
                   contributor=contributor,
                   modified=modified,
                   projectIri=projectIri,
                   projectShortName=projectShortName,
                   label=label,
                   namespaceIri=namespaceIri,
                   comment=comment,
                   projectStart=projectStart,
                   projectEnd=projectEnd)

    @staticmethod
    def search(con: IConnection,
               label: Xsd_string | str | None = None,
               comment: Xsd_string | str | None = None) -> list[Iri]:
        """
        Search for a given project. If no label or comment is given, all existing projects are returned. If both
        a search term for the label and comment are given, they will be combined by *AND*.
        :param con: Valid Connection object
        :type con: IConnection
        :param label: A string to search for in the labels. The search will check if the label of a project
        **contains** the string given here.
        :type label: str
        :param comment: A string to search for in the comments. The search will check if the comment of a project
        **contains** the string given here
        :type comment: str
        :return: List of IRIs matching the search criteria (AnyIRI | QName)
        :rtype: List[Iri]
        :raises OldapErrorNotFound: If the project does not exist
        """
        label = Xsd_string(label)
        comment = Xsd_string(comment)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += 'SELECT DISTINCT ?project\n'
        sparql += 'FROM oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '   ?project a oldap:Project .\n'
        if label:
            sparql += '   ?project rdfs:label ?label .\n'
            sparql += f'   FILTER(CONTAINS(STR(?label), "{Xsd_string.escaping(label.value)}"))\n'
        if comment:
            sparql += '   ?project rdfs:comment ?comment .\n'
            sparql += f'   FILTER(CONTAINS(STR(?comment), "{Xsd_string.escaping(comment.value)}"))\n'
        sparql += '}\n'
        # sparql += f"""
        # SELECT DISTINCT ?project
        # FROM oldap:admin
        # WHERE {{
        #     ?project rdfs:label ?label
        #     FILTER(STRSTARTS(?label, "{label}"))
        # }}
        # """
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        projects: list[Iri] = []
        if len(res) > 0:
            for r in res:
                projects.append(r['project'])
        return projects

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Create a new project in the triple store
        :param indent: Start indent level for generated SPARQL (debugging)
        :type indent: int
        :param indent_inc: Indent increment
        :type indent_inc: int
        :return: None
        :raises OldapErrorAlreadyExists: If a project with the projectIri already exists
        :raises OldapError: All other errors
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        indent: int = 0
        indent_inc: int = 4

        context = Context(name=self._con.context_name)

        sparql1 = context.sparql_context
        sparql1 += f"""
        SELECT ?project
        FROM oldap:admin
        WHERE {{
            ?project a oldap:Project .
            FILTER(?project = {self.projectIri.toRdf})
        }}
        """

        blank = ''
        sparql2 = context.sparql_context
        sparql2 += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql2 += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{'
        sparql2 += f'\n{blank:{(indent + 2) * indent_inc}}{self.projectIri.toRdf} a oldap:Project'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}oldap:projectShortName {self.projectShortName.toRdf}'
        if self.label:
            sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}rdfs:label {self.label.toRdf}'
        if self.comment:
            sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}rdfs:comment {self.comment.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}oldap:namespaceIri {self.namespaceIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}oldap:projectStart {self.projectStart.toRdf}'
        if self.projectEnd is not None:
            sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}oldap:projectEnd {self.projectEnd.toRdf}'
        sparql2 += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql2 += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            jsonobj = self._con.transaction_query(sparql1)
        except OldapError:
            self._con.transaction_abort()
            raise
        res = QueryProcessor(context, jsonobj)
        if len(res) > 0:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A Project with a projectIri "{self.projectIri}" already exists')

        try:
            self._con.transaction_update(sparql2)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__created = timestamp
        self.__creator = self._con.userIri
        self.__modified = timestamp
        self.__contributor = self._con.userIri
        context[self.__attributes[ProjectAttr.PROJECT_SHORTNAME]] = self.__attributes[ProjectAttr.NAMESPACE_IRI]

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Write all the modifications that were applied to the project instqnce to the triple store.
        :param indent: Starting indent for SPARQL queries [Only used for debbugging purposes]
        :type indent: int
        :param indent_inc: Indent increment for SPARQL queries [Only used for debbugging purposes]
        :type indent_inc: int
        :return: None
        :Raises: OldapErrorNoPermission: No permission for operation
        :Raises: OldapErrorUpdateFailed: Update failed
        :Raises: Oldap Error: Other Internal error
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        blank = ''
        sparql_list = []
        for field, change in self.__changeset.items():
            if field == ProjectAttr.LABEL or field == ProjectAttr.COMMENT:
                if change.action == Action.MODIFY:
                    sparql_list.extend(self.__attributes[field].update(graph=Xsd_QName('oldap:admin'),
                                                                       subject=self.projectIri,
                                                                       subjectvar='?project',
                                                                       field=Xsd_QName(field.value)))
                if change.action == Action.DELETE or change.action == Action.REPLACE:
                    sparql = self.__changeset[field].old_value.delete(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.projectIri,
                                                                      field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                if change.action == Action.CREATE or change.action == Action.REPLACE:
                    sparql = self.__attributes[field].create(graph=Xsd_QName('oldap:admin'),
                                                             subject=self.projectIri,
                                                             field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                continue
            sparql = f'{blank:{indent * indent_inc}}# Project field "{field.value}" with action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {self.__attributes[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.projectIri.toRdf} as ?project)\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)

        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri, self.modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri)
        except OldapError:
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed("Update failed! Timestamp does not match")
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__modified = timestamp
        self.__contributor = self._con.userIri  # TODO: move creator, created etc. to Model!

    def delete(self) -> None:
        """
        Delete the given user from the triplestore
        :return: None
        :raises OldapErrorNoPermission: No permission for operation
        :raises OldapError: generic internal error
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        #
        # TODO: Check if project as any datamodel and/or data. Decline the deletion if this is the case
        #
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            {self.projectIri.toRdf} a oldap:Project .
            {self.projectIri.toRdf} ?prop ?val .
        }} 
        """
        # TODO: use transaction for error handling
        self._con.update_query(sparql)




import math
import re
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.floatingpoint import FloatingPoint
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_float(FloatingPoint):
    """
    Implements the XML Schema [xsd:float](https://www.w3.org/TR/xmlschema11-2/#float) datatype
    """

    def __init__(self, value: Self | float | str, validate: bool = True):
        """
        Constructor for Xsd_float class
        :param value: A Xsd_float value, a floating point or a string containing a valid float value
        :type value: Xsd_float | float | str
        :raises OldapErrorValue: If the value is not a float or a string containing a valid float value
        """
        if isinstance(value, str):
            if validate:
                if not re.match("^([-+]?(\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?|[Nn]a[Nn]|[-+]?(inf|INF))$", value):
                    raise OldapErrorValue(f'"{value}" is not a xsd:float.')
            value = float(value)
        super().__init__(value)

    @property
    def toRdf(self) -> str:
        return self._toRdf('xsd:float')
